# Copyright(c) 2016, Oracle and/or its affiliates.  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Exception definitions."""
import kolla_cli.i18n as u


class ClientException(Exception):
    """KollaClient Base Class Exception"""
    def __init__(self, message, *args):
        if not message:
            message = u._('An unknown exception occurred.')
        super(ClientException, self).__init__(message, *args)


class NotInInventory(ClientException):
    """Not in inventory exception"""
    def __init__(self, obj_type, obj_names, *args):
        if isinstance(obj_names, list):
            # list of names
            invalid_objs = ''
            comma = ''
            for obj_name in obj_names:
                invalid_objs = ''.join([invalid_objs, comma, obj_name])
                comma = ','
        else:
            # single object name
            invalid_objs = obj_names
        message = (u._('{type} ({objs}) does not exist.')
                   .format(type=obj_type, objs=invalid_objs))
        super(NotInInventory, self).__init__(message, *args)


class HostError(ClientException):
    pass


class HostsSshCheckError(ClientException):
    """Host failed its ssh check"""
    def __init__(self, hostnames, *args):
        failed_hosts = ''
        comma = ''
        for hostname in hostnames:
            failed_hosts = ''.join([failed_hosts, comma, hostname])
            comma = ','
        message = (u._('Host(s) ssh check failed: {hosts}')
                   .format(hosts=failed_hosts))
        super(HostsSshCheckError, self).__init__(message, *args)


class InvalidArgument(ClientException):
    """Invalid argument"""
    pass


class InvalidConfiguration(ClientException):
    """Invalid configuration"""
    pass


class FailedOperation(ClientException):
    pass


class MissingArgument(ClientException):
    """Missing argument"""
    def __init__(self, argname, *args):
        message = (u._('Argument is missing: {name}')
                   .format(name=argname))
        super(MissingArgument, self).__init__(message, *args)



# Copyright(c) 2016, Oracle and/or its affiliates.  All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#
import os

from kolla_cli.common.service import Service
from kolla_cli.common.utils import get_kolla_ansible_home


class AnsibleInventory(object):
    """AnsibleInventory helper class

    This class parses an ansible inventory file and provides an
    easier to use way to represent that file.
    """
    def __init__(self, inventory_path=None):
        self.groups = []
        self.services = {}
        self.group_vars = {}
        self.host_vars = {}

        self._load(inventory_path)

    def add_service(self, servicename):
        if servicename not in self.services:
            service = Service(servicename)
            self.services[servicename] = service
        return self.services[servicename]

    def add_group(self, groupname):
        if groupname not in self.groups:
            self.groups.append(groupname)

    def _load(self, inventory_path):
        """load an ansible inventory file

        Note: This assumes that there will be a blank line between each
        section:

        # Mistral
        [mistral-api:children]
        mistral

        [mistral-executor:children]
        mistral
        """
        if not inventory_path:
            inventory_path = os.path.join(
                get_kolla_ansible_home(), 'ansible', 'inventory',
                'all-in-one')
        with open(inventory_path, 'r') as ain1:
            ain1_inv = ain1.read()

        lines = [x for x in ain1_inv.split('\n') if not x.startswith('#')]
        for i in range(0, len(lines)):
            line = lines[i]
            if not line.startswith('['):
                continue
            line.strip()
            if ':vars' in line:
                # this is defining a set of group vars in the next set
                # of lines.
                groupname = line.split(':')[0]
                i = self._process_group_vars(groupname, lines, i)
                continue
            if ':children' not in line:
                groupname = line[1:len(line) - 1]
                self.add_group(groupname)
                continue

            servicename = line.split(':children')[0]
            servicename = servicename[1:]
            service = self.add_service(servicename)

            # next lines will be parents or groups for service found above
            has_parents_or_groups = False
            while True:
                i += 1
                line = lines[i]
                parent_or_group = line.strip()
                if parent_or_group.startswith('#'):
                    # comment line, skip
                    continue
                if not parent_or_group:
                    # blank line, done processing parents
                    # if a service has no parent or group associations
                    # we infer that it is not supported and is filtered
                    # from being shown to the client
                    service.set_supported(has_parents_or_groups)
                    break
                if parent_or_group in self.groups:
                    service.add_groupname(parent_or_group)
                    has_parents_or_groups = True
                else:
                    service.add_parentname(parent_or_group)
                    has_parents_or_groups = True

        for _, service in self.services.items():
            for parentname in service.get_parentnames():
                parent = self.services[parentname]
                if parent:
                    parent.add_childname(service.name)

    def _process_group_vars(self, groupname, lines, i):
        # currently group vars are ignored
        while True:
            i += 1
            line = lines[i].strip()
            if not line:
                break



# Copyright(c) 2018, Oracle and/or its affiliates.  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from kolla_cli.api.job import Job
from kolla_cli.common.ansible.actions import KollaAction
from kolla_cli.common.utils import check_arg
import kolla_cli.i18n as u


class CertificateApi(object):

    @staticmethod
    def certificate_init(verbose_level=1):
        """Certificate Init.

        Creates a self-signed certificate for secure TLS communication.

        :param verbose_level: the higher the number, the more verbose
        :type verbose_level: integer
        :return: Job object
        :rtype: Job
        """
        check_arg(verbose_level, u._('Verbose level'), int)
        action = KollaAction(verbose_level=verbose_level,
                             playbook_name='certificates.yml')
        ansible_job = action.certificate_init()
        return Job(ansible_job)



# Copyright(c) 2016, Oracle and/or its affiliates.  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import traceback

from cliff.command import Command
from cliff.lister import Lister

from kolla_cli.api.client import ClientApi
from kolla_cli.api.exceptions import ClientException
from kolla_cli.commands.exceptions import CommandError
from kolla_cli.common.utils import convert_lists_to_string
import kolla_cli.i18n as u

CLIENT = ClientApi()


class ServiceAddGroup(Command):
    """Add group to service.

    Associated the service to a group. If this is a sub-service,
    the inherit flag will be cleared.
    """

    def get_parser(self, prog_name):
        parser = super(ServiceAddGroup, self).get_parser(prog_name)
        parser.add_argument('servicename', metavar='<servicename>',
                            help=u._('Service name'))
        parser.add_argument('groupname', metavar='<groupname>',
                            help=u._('Group name'))
        return parser

    def take_action(self, parsed_args):
        try:
            groupname = parsed_args.groupname.strip()
            servicename = parsed_args.servicename.strip()

            group = CLIENT.group_get([groupname])[0]
            group.add_service(servicename)

        except ClientException as e:
            raise CommandError(str(e))
        except Exception:
            raise Exception(traceback.format_exc())


class ServiceRemoveGroup(Command):
    """Remove group from service."""

    def get_parser(self, prog_name):
        parser = super(ServiceRemoveGroup, self).get_parser(prog_name)
        parser.add_argument('servicename', metavar='<servicename>',
                            help=u._('Service name'))
        parser.add_argument('groupname', metavar='<groupname>',
                            help=u._('Group name'))
        return parser

    def take_action(self, parsed_args):
        try:
            groupname = parsed_args.groupname.strip()
            servicename = parsed_args.servicename.strip()

            group = CLIENT.group_get([groupname])[0]
            group.remove_service(servicename)

        except ClientException as e:
            raise CommandError(str(e))
        except Exception:
            raise Exception(traceback.format_exc())


class ServiceListGroups(Lister):
    """List services and their groups."""

    def take_action(self, parsed_args):
        try:
            data = [('', '')]
            services = CLIENT.service_get_all()
            if services:
                data = []
                for service in services:
                    groupnames = sorted(service.get_groups())
                    data.append((service.name, groupnames))

            data = convert_lists_to_string(data, parsed_args)
            return (u._('Service'), u._('Groups')), sorted(data)
        except ClientException as e:
            raise CommandError(str(e))
        except Exception:
            raise Exception(traceback.format_exc())


class ServiceList(Lister):
    """List services and their sub-services."""

    def take_action(self, parsed_args):
        try:
            data = [('', '')]
            services = CLIENT.service_get_all()
            if services:
                data = []
                for service in services:
                    data.append((service.name,
                                 sorted(service.get_children())))

            data = convert_lists_to_string(data, parsed_args)
            return ((u._('Service'), u._('Children')), sorted(data))

        except ClientException as e:
            raise CommandError(str(e))
        except Exception:
            raise Exception(traceback.format_exc())



# Copyright(c) 2016, Oracle and/or its affiliates.  All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#
import logging
import os
import shutil
import subprocess
import sys
import testtools

import kolla_cli.common.utils as utils

from copy import copy
from shutil import copyfile

from kolla_cli.api.client import ClientApi
from kolla_cli.api.exceptions import InvalidArgument

CLIENT = ClientApi()
CLIENT.enable_console_logging(logging.DEBUG)

ARG_LIST = {
    bool: False,
    list: [1, 2, 3],
    str: 'qwerty',
    dict: {'a': 1},
    int: 0,
    }

KOLLACLI_CMD_PATH = 'kolla_cli/shell.py'


class KollaCliTest(testtools.TestCase):

    saved_kolla_etc = ''
    log = logging.getLogger(__name__)

    def setUp(self):
        super(KollaCliTest, self).setUp()

        logging.basicConfig(stream=sys.stderr)
        self.log.setLevel(logging.DEBUG)
        self.log.info('\nStarting test: %s ***********************************'
                      % self._testMethodName)

        # switch to test path
        etc_path = utils.get_kolla_cli_etc()
        self.log.debug('etc for tests: %s' % etc_path)

        self._save_config()

    def tearDown(self):
        super(KollaCliTest, self).tearDown()
        self._restore_config()

    def run_cli_cmd(self, cmd, expect_error=False):
        full_cmd = ('python %s %s' % (KOLLACLI_CMD_PATH, cmd))
        self.log.debug('running command: %s' % cmd)
        (retval, msg) = self.run_command(full_cmd)

        if not expect_error:
            self.assertEqual(0, retval, ('command failed: (%s), cmd: %s'
                                         % (msg, full_cmd)))
        return msg

    def run_command(self, cmd):
        """run bash command

        return (retval, msg)
        """
        # self.log.debug('run cmd: %s' % cmd)
        msg = ''
        process = subprocess.Popen(cmd,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE,
                                   shell=True)
        (out, err) = process.communicate()
        retval = process.returncode

        # the py dev debugger adds a string at the line start, remove it
        if err:
            msg = utils.safe_decode(err)
        if out:
            msg = msg + '\n' + utils.safe_decode(out)
        if msg.startswith('pydev debugger'):
            msg = msg.split('\n', 1)[1]
        return (retval, msg)

    def check_types(self, method, expected_types):
        # expected type is a list:
        args = []
        for arg_type in expected_types:
            args.append((arg_type, ARG_LIST[arg_type]))
        for i in range(0, len(args)):
            arg_type, _ = args[i]
            for new_arg in ARG_LIST.values():
                new_args = copy(args)
                if isinstance(new_arg, arg_type):
                    # this new type is the correct arg type, skip
                    continue
                # substitute a valid type for an invalid one
                new_args[i] = new_arg
                self._check_invalid_arg(method, new_args)

    def _check_invalid_arg(self, method, args):
        assert(len(args) <= 5)
        try:
            if len(args) == 1:
                method(args[0])
            elif len(args) == 2:
                method(args[0], args[1])
            elif len(args) == 3:
                method(args[0], args[1], args[2])
            elif len(args) == 4:
                method(args[0], args[1], args[2], args[3])
            elif len(args) == 5:
                method(args[0], args[1], args[2], args[3], args[4])
        except InvalidArgument:
            # success
            return
        except Exception as e:
            self.assertTrue(False,
                            'method: %s, arg: %s ' % (method, args) +
                            'failed with the wrong exception: '
                            '%s' % str(e))
        self.assertTrue(False, 'method: %s, arg: %s did not fail'
                        % (method, args))

    # PRIVATE FUNCTIONS ----------------------------------------------------
    def _save_config(self):
        """save config"""
        # save inventory
        src_path = os.path.join(utils.get_kolla_cli_etc(),
                                'ansible', 'inventory.json')
        dst_path = os.path.join('/tmp', 'inventory.json.utest.save')
        copyfile(src_path, dst_path)

        # save group vars
        ansible_dir = os.path.join(utils.get_kolla_ansible_home(), 'ansible')
        groupdir = os.path.join(ansible_dir, 'group_vars')
        self._save_dir(groupdir)

        # save host vars
        hostdir = os.path.join(ansible_dir, 'host_vars')
        self._save_dir(hostdir)

    def _restore_config(self):
        """restore config"""
        # restore inventory
        dst_path = os.path.join(utils.get_kolla_cli_etc(),
                                'ansible', 'inventory.json')
        src_path = os.path.join('/tmp', 'inventory.json.utest.save')
        copyfile(src_path, dst_path)

        # restore group vars
        ansible_dir = os.path.join(utils.get_kolla_ansible_home(), 'ansible')
        groupdir = os.path.join(ansible_dir, 'group_vars')
        self._restore_dir(groupdir)

        # restore host vars
        hostdir = os.path.join(ansible_dir, 'host_vars')
        self._restore_dir(hostdir)

    def _save_dir(self, src_dir):
        if not os.path.exists(src_dir):
            return
        dirname = os.path.basename(src_dir)
        save_dir = os.path.join('/tmp', dirname + '.utest.save')
        if os.path.exists(save_dir):
            shutil.rmtree(save_dir)
        os.mkdir(save_dir)
        fnames = os.listdir(src_dir)
        for fname in fnames:
            src_path = os.path.join(src_dir, fname)
            dst_path = os.path.join(save_dir, fname)
            copyfile(src_path, dst_path)

    def _restore_dir(self, dst_dir):
        # we do not have privs to write these files
        ignore_list = ['all.yml']

        dirname = os.path.basename(dst_dir)
        save_dir = os.path.join('/tmp', dirname + '.utest.save')
        sv_fnames = os.listdir(save_dir)
        fnames = os.listdir(dst_dir)
        # remove any new var files created by tests
        for fname in fnames:
            if fname not in sv_fnames:
                os.remove(os.path.join(dst_dir, fname))
        # restore saved files
        for sv_fname in sv_fnames:
            if sv_fname in ignore_list:
                continue
            src_path = os.path.join(save_dir, sv_fname)
            dst_path = os.path.join(dst_dir, sv_fname)
            copyfile(src_path, dst_path)



# Copyright(c) 2016, Oracle and/or its affiliates.  All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#
import os
import unittest

from kolla_cli.api import client
from kolla_cli.common.utils import get_kolla_etc
from kolla_cli.tests.functional.common import KollaCliTest

CLIENT = client.ClientApi()

PUBLIC_KEY = (
    'ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAACAQCqusDp5jpkbng3sRue8gZV6/PCQp9'
    'ogUd5/OZ3sh9VgdaigHoYUfXZElTZLlkL71tD9WZJr69PDwmG/nE4quba8rLcDY2wC0'
    'qjq+r06ExhlRu4ivy7OxT29s8FSe8Uht9Pz8ahnXxddLF55yTbC81XrSXDBFc6Nnogz'
    '+g6GgXVKtwTkm5g3K+qix5zVECu8zzawBR/s+v0dkDxKwSY8XOG6JZlMUndaDaikZZi'
    'qp8KAOJpajM77aCfDkY3VZGFBCJiEGLVDhFrtXuBI9I0YzX4j9pZZWpSzkM/FwlPjDR'
    'SW1C9MAAFLoEQTN4j1Z5hkDNXDsr49wJBi+jjQ0FPMMvfJktrznRuO2fUa9W2iilOrv'
    '1PyrknssmW1iYXiWJ5Bq8A9sKE1r7Nbdjhcjskp77X57tNjtarRUcj3FqGjC8pv+k92'
    '9Y+FvkXbjpBsHpdMFh8BlM+EnwnsjkiQpmjLv8bpeeQooLyQQmZn94zY73bbGrsjzXe'
    'OhOTDnKAS14hxCnBlEbudHB4erp/5Nj+A8UVAT0KXPM+mkDrum/dsvV0wnvBicAVt/a'
    'tmkwDKJqXDmj4elNe8/jTXSYHpTDo29xtcGpka9AtWarmnt8QkRuieD1xSXsEUQswjq'
    'aQD2ikitKt/hEyCmT+7fy4yYKK35kukUj5qV85A8O/hOYf5vFjtRw==')

PRIVATE_KEY = (
    '-----BEGIN RSA PRIVATE KEY-----\n'
    'MIIJKAIBAAKCAgEAqrrA6eY6ZG54N7EbnvIGVevzwkKfaIFHefzmd7IfVYHWooB6\n'
    'GFH12RJU2S5ZC+9bQ/VmSa+vTw8Jhv5xOKrm2vKy3A2NsAtKo6vq9OhMYZUbuIr8\n'
    'uzsU9vbPBUnvFIbfT8/GoZ18XXSxeeck2wvNV60lwwRXOjZ6IPoOhoF1SrcE5JuY\n'
    'Nyvqosec1RArvM82sAUf7Pr9HZA8SsEmPFzhuiWZTFJ3Wg2opGWYqqfCgDiaWozO\n'
    '+2gnw5GN1WRhQQiYhBi1Q4Ra7V7gSPSNGM1+I/aWWVqUs5DPxcJT4w0UltQvTAAB\n'
    'S6BEEzeI9WeYZAzVw7K+PcCQYvo40NBTzDL3yZLa850bjtn1GvVtoopTq79T8q5J\n'
    '7LJltYmF4lieQavAPbChNa+zW3Y4XI7JKe+1+e7TY7Wq0VHI9xahowvKb/pPdvWP\n'
    'hb5F246QbB6XTBYfAZTPhJ8J7I5IkKZoy7/G6XnkKKC8kEJmZ/eM2O922xq7I813\n'
    'joTkw5ygEteIcQpwZRG7nRweHq6f+TY/gPFFQE9ClzzPppA67pv3bL1dMJ7wYnAF\n'
    '69VedCYMSoYIHpcN80w9it/6Cfm8niAy3v9e0icSVEsvkzcV6eFjLggY1DQ9WBPN\n'
    'MR4LKGNDuxEWeZAQi+A6Ejclx1KKBhL/E4SNj3ev4/5glaMjzSIUpA4415o=\n'
    '-----END RSA PRIVATE KEY-----'
    )


class TestFunctional(KollaCliTest):

    def test_password_set_clear(self):
        # test list
        msg = self.run_cli_cmd('password list')
        key = 'database_password'
        value = '-'
        ok = self._password_value_exists(key, value, msg)
        self.assertTrue(ok, 'list failed. Password (%s/%s) not in output: %s'
                        % (key, value, msg))

        # test setting empty password
        self.run_cli_cmd('password set %s --insecure' % key)
        msg = self.run_cli_cmd('password list')
        ok = self._password_value_exists(key, '-', msg)
        self.assertTrue(ok, 'set empty password failed. Password ' +
                        '(%s/-) not in output: %s' %
                        (key, msg))

        # test setting None password
        CLIENT.password_set(key, None)
        msg = self.run_cli_cmd('password list')
        ok = self._password_value_exists(key, '-', msg)
        self.assertTrue(ok, 'set None password failed. Password ' +
                        '(%s/-) not in output: %s' %
                        (key, msg))

        # test clear
        key = 'database_password'
        value = '-'
        self.run_cli_cmd('password clear %s' % key)
        msg = self.run_cli_cmd('password list')
        ok = self._password_value_exists(key, value, msg)
        self.assertTrue(ok, 'clear password failed. Password ' +
                        '(%s/%s) not in output: %s' %
                        (key, value, msg))

        # test setting an ssh key
        key = 'nova_ssh_key'
        CLIENT.password_set_sshkey(key, PRIVATE_KEY, PUBLIC_KEY)
        keynames = CLIENT.password_get_names()
        self.assertIn(key, keynames, 'ssh key not in passwords')

        # test modify non-ssh password
        key = 'database_password'
        value = '-'
        self.run_cli_cmd('password set %s --insecure %s' % (key, value))
        msg = self.run_cli_cmd('password list')
        ok = self._password_value_exists(key, value, msg)
        self.assertTrue(ok, 'set modify password failed. Password ' +
                        '(%s/%s) not in output: %s' %
                        (key, value, msg))

        # test to make sure that saves / loads aren't doing something
        # bad to the password file size
        CLIENT.password_clear(key)
        # snapshot file size with key cleared
        password_file_path = os.path.join(get_kolla_etc(), 'passwords.yml')
        size_start = os.path.getsize(password_file_path)
        # set and clear password
        CLIENT.password_set(key, value)
        CLIENT.password_clear(key)
        size_end = os.path.getsize(password_file_path)
        self.assertEqual(size_start, size_end, 'password file size changed ' +
                         'during set/clear (%s/%s)' % (size_start, size_end))

        # make sure to end the test with the password init, as some other
        # non-password related tests require that all passwords in the file
        # be populated
        CLIENT.password_init()

    def _password_value_exists(self, key, value, cli_output):
        """Verify cli data against model data"""
        # check for any host in cli output that shouldn't be there
        cli_lines = cli_output.split('\n')
        for cli_line in cli_lines:
            if key in cli_line and value in cli_line:
                return True
        return False


if __name__ == '__main__':
    unittest.main()



# Copyright(c) 2016, Oracle and/or its affiliates.  All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#


class Host(object):
    class_version = 1

    def __init__(self, hostname):
        self.name = hostname
        self.alias = ''
        self.is_mgmt = False
        self.hypervisor = ''
        self.vars = {}
        self.version = self.__class__.class_version

    def get_vars(self):
        return self.vars.copy()

    def set_var(self, name, value):
        self.vars[name] = value

    def upgrade(self):
        pass



# Copyright(c) 2016, Oracle and/or its affiliates.  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from copy import copy

from kolla_cli.common.inventory import Inventory
from kolla_cli.common.utils import check_arg
from kolla_cli.common.utils import safe_decode
import kolla_cli.i18n as u

MYPY = False
if MYPY:
    from typing import List  # noqa


class ServiceApi(object):

    def service_get_all(self):
        # type: () -> List[Service]
        """Get all services in the inventory

        :return: services
        :rtype: List of Service objects
        """
        return self._get_services([], get_all=True)

    def service_get(self, servicenames):
        # type: (List[str]) -> List[Service]
        """Get selected services in the inventory

        :param servicenames: names of services to be read
        :type servicenames: list of strings
        :return: services
        :rtype: list of Service objects
        """
        check_arg(servicenames, u._('Service names'), list)
        servicenames = safe_decode(servicenames)
        return self._get_services(servicenames)

    def _get_services(self, servicenames, get_all=False):
        # type: (List[str], bool) -> List[Service]
        services = []
        inventory = Inventory.load()

        if get_all:
            inv_services = inventory.get_services(client_filter=True)
            for inv_service in inv_services:
                service = Service(inv_service.name,
                                  inv_service.get_parentnames(),
                                  inv_service.get_childnames(),
                                  inv_service.get_groupnames())
                services.append(service)
        else:
            inventory.validate_servicenames(servicenames, client_filter=True)

            for servicename in servicenames:
                inv_service = inventory.get_service(servicename,
                                                    client_filter=True)
                if inv_service:
                    service = Service(inv_service.name,
                                      inv_service.get_parentnames(),
                                      inv_service.get_childnames(),
                                      inv_service.get_groupnames())
                services.append(service)
        return services


class Service(object):
    """Service

    A service is one of the services available in openstack-kolla-ansible.

    For example, this would be how the murano services would be
    represented:

    - murano
        - parentnames: []
        - childnames: [murano-api, murano-engine]
    - murano-api
        - parentnames: [murano]
        - childnames: []
    - murano-engine
        - parentnames: [murano]
        - childnames: []
    """

    def __init__(self, servicename, parentnames=[],
                 childnames=[], groupnames=[]):
        # type: (str, List[str], List[str], List[str]) -> None
        self.name = servicename
        self._parentnames = parentnames
        self._childnames = childnames
        self._groupnames = groupnames

    def get_name(self):
        # type: () -> str
        """Get name

        :return: service name
        :rtype: string
        """
        return self.name

    def get_parents(self):
        # type: () -> List[str]
        """Get name of parent services

        :return: parent service names
        :rtype: string
        """
        return copy(self._parentnames)

    def get_children(self):
        # type: () -> List[str]
        """Get names of the child services

        :return: child names
        :rtype: list of strings
        """
        return copy(self._childnames)

    def get_groups(self):
        # type: () -> List[str]
        """Get names of the groups

        :return: group names
        :rtype: list of strings

        Note: If the groups associated with this service change after this
        service is fetched, the service must be re-fetched to reflect those
        changes.
        """
        return copy(self._groupnames)



# Copyright(c) 2016, Oracle and/or its affiliates.  All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#
from kolla_cli.tests.functional.common import KollaCliTest

import json
import os
import unittest

from kolla_cli.common.utils import get_group_vars_dir
from kolla_cli.common.utils import get_host_vars_dir
from kolla_cli.common.utils import get_kolla_ansible_home

from kolla_cli.common.inventory import Inventory


class TestFunctional(KollaCliTest):

    def test_properties(self):
        # test global properties
        self._properties_test()

        # test single group vars
        group = 'prop_test_group1'
        self.run_cli_cmd('group add %s' % group)
        self._properties_test(groups=['control'])

        # test single host vars
        host = 'prop_test_host1'
        self.run_cli_cmd('host add %s' % host)
        self._properties_test(hosts=[host])

        # test multiple group vars
        groups = [group]
        group = 'prop_test_group2'
        groups.append(group)
        self.run_cli_cmd('group add %s' % group)
        self._properties_test(groups=groups)

        # test multiple host vars
        hosts = [host]
        host = 'prop_test_host2'
        hosts.append(host)
        self.run_cli_cmd('host add %s' % host)
        self._properties_test(hosts=hosts)

        # test all group vars
        self._properties_test(groups=['all'])

        # test all host vars
        self._properties_test(hosts=['all'])

        # test property override output
        ovr_key = 'enable_haproxy'
        ovr_value = 'no'

        # clear property values before test
        self.run_cli_cmd('property clear %s' % ovr_key)
        self.run_cli_cmd('property clear %s --host=all' % ovr_key)
        self.run_cli_cmd('property clear %s --group=all' % ovr_key)

        # global property override test
        self.run_cli_cmd('property set %s %s' % (ovr_key, ovr_value))
        json_str = self.run_cli_cmd('property list -f json')
        msg = self._override_test(json_str, ovr_key, ovr_value, '*--')
        self.assertEqual(msg, '', 'override check failed: %s' % msg)

        # host property override test
        self.run_cli_cmd('property set %s %s --host=%s' %
                         (ovr_key, ovr_value, host))
        json_str = self.run_cli_cmd('property list -f json --host=%s' % host)
        msg = self._override_test(json_str, ovr_key,
                                  ovr_value, '*-H', host=host)
        self.assertEqual(msg, '', 'host override check failed: %s' % msg)

        # group property override test
        self.run_cli_cmd('property set %s %s --group=%s' %
                         (ovr_key, ovr_value, group))
        json_str = self.run_cli_cmd('property list -f json --group=%s' % group)
        msg = self._override_test(json_str, ovr_key,
                                  ovr_value, '*GH', group=group)
        self.assertEqual(msg, '', 'group override check failed: %s' % msg)

        # check that group_var files are deleted
        # when groups are deleted
        for group in groups:
            path = os.path.join(get_group_vars_dir(), group)
            self.assertTrue(os.path.exists(path))
            self.run_cli_cmd('group remove %s' % group)
            self.assertFalse(os.path.exists(path))

        # check that host_var files are deleted
        # when hosts are deleted
        for host in hosts:
            path = os.path.join(get_host_vars_dir(), host)
            self.assertTrue(os.path.exists(path))
            self.run_cli_cmd('host remove %s' % host)
            self.assertFalse(os.path.exists(path))

    def _properties_test(self, groups=[], hosts=[]):
        switch = ''
        if groups:
            switch = '--groups'
            dir_name = 'group_vars'
        elif hosts:
            switch = '--hosts'
            dir_name = 'host_vars'

        key = 'TeStKeY'
        value = 'TeStVaLuE:123:abc'

        # initialize keys
        targets_csv = ''
        targets = groups + hosts
        if 'all' in groups:
            inv = Inventory.load()
            targets = inv.get_groupnames()
            targets_csv = 'all'
        elif 'all' in hosts:
            inv = Inventory.load()
            targets = inv.get_hostnames()
            targets_csv = 'all'

        comma = ''
        sizes = {}  # key = path, value = [size1, size2, etc]
        for target in targets:
            self.run_cli_cmd('property clear %s %s %s'
                             % (switch, target, key))
            if targets_csv != 'all':
                targets_csv += comma + target
                comma = ','
            path = os.path.join(get_kolla_ansible_home(),
                                'ansible', dir_name, target)
            sizes[path] = [os.path.getsize(path)]
        if not switch:
            self.run_cli_cmd('property clear %s' % key)
            path = os.path.join(get_kolla_ansible_home(),
                                'ansible/group_vars/__GLOBAL__')
            sizes[path] = [os.path.getsize(path)]

        # test append
        self.run_cli_cmd('property set %s %s %s %s'
                         % (switch, targets_csv, key, value))
        if switch:
            msg = self.run_cli_cmd('property list -f json %s all'
                                   % (switch))
        else:
            msg = self.run_cli_cmd('property list -f json')
        err_msg = self._check_property_values(key, value, msg, targets)
        self.assertEqual(err_msg, '',
                         'set failed property not in output: %s, %s (%s %s)'
                         % (key, value, switch, targets_csv))

        bad_path = self._is_size_ok(sizes, 0, '<', 1)
        self.assertIsNone(bad_path, 'Size of file %s did not ' % bad_path +
                          'increase after append (%s %s)'
                          % (switch, targets_csv))

        # test modify existing
        value += '2'
        self.run_cli_cmd('property set %s %s %s %s'
                         % (switch, targets_csv, key, value))
        msg = self.run_cli_cmd('property list --all -f json %s %s'
                               % (switch, targets_csv))
        err_msg = self._check_property_values(key, value, msg, targets)
        self.assertEqual(err_msg, '',
                         'set failed property not in output: %s, %s (%s %s)'
                         % (key, value, switch, targets_csv))
        bad_path = self._is_size_ok(sizes, 1, '<', 2)
        self.assertIsNone(bad_path, 'Size of file %s did not ' % bad_path +
                          'increase after modify (%s %s)'
                          % (switch, targets_csv))

        # test clear
        self.run_cli_cmd('property clear %s %s %s'
                         % (switch, targets_csv, key))
        msg = self.run_cli_cmd('property list --long -f json %s %s'
                               % (switch, targets_csv))
        err_msg = self._check_property_values(key, value, msg, targets)
        self.assertTrue('missing' in err_msg,
                        'clear failed, property still in output: ' +
                        '%s, %s (%s %s)'
                        % (key, value, switch, targets_csv))
        bad_path = self._is_size_ok(sizes, 0, '=', 3)
        self.assertIsNone(bad_path, 'Size of file %s is ' % bad_path +
                          'different from initial size '
                          '(%s %s %s)'
                          % (switch, targets_csv, str(sizes)))

        # test setting empty string
        value = '""'
        self.run_cli_cmd('property set %s %s %s %s'
                         % (switch, targets_csv, key, value))
        msg = self.run_cli_cmd('property list --all -f json %s %s'
                               % (switch, targets_csv))
        err_msg = self._check_property_values(key, value, msg, targets)
        self.assertTrue('missing' in err_msg,
                        'clear failed, property still in output: ' +
                        '%s, %s (%s %s)'
                        % (key, value, switch, targets_csv))
        self.run_cli_cmd('property clear %s %s %s'
                         % (switch, targets_csv, key))

    def _check_property_values(self, key, value, json_str,
                               targets=[]):
        """Verify cli data against model data"""
        error_msg = ''
        props = json.loads(json_str.strip())
        if not targets:
            # simple property check
            ok = False
            for prop in props:
                if (prop['Property Name'] == key and
                        prop['Property Value'] == value):
                    ok = True
            if not ok:
                error_msg = '%s:%s is missing in __GLOBAL__'
        else:
            target_map = {}
            for target in targets:
                target_map[target] = 'missing'
                for prop in props:
                    if 'Group' in prop and prop['Group'] == target:
                        if (prop['Property Name'] == key and
                                prop['Property Value'] == value):
                            target_map[target] = 'ok'
                    elif 'Host' in prop and prop['Host'] == target:
                        if (prop['Property Name'] == key and
                                prop['Property Value'] == value):
                            target_map[target] = 'ok'
            for target, state in target_map.items():
                if state == 'missing':
                    error_msg += ('%s:%s is missing in %s\n output:%s\n'
                                  % (key, value, target, json_str))
        return error_msg

    def _is_size_ok(self, sizes, idx0, comparator, idx1):
        bad_path = None
        for path, path_sizes in sizes.items():
            if idx1 > len(path_sizes) - 1:
                # get new sizes
                sizes[path].append(os.path.getsize(path))
            if comparator == '=':
                if sizes[path][idx0] != sizes[path][idx1]:
                    bad_path = path
                    break
            elif comparator == '<':
                if sizes[path][idx0] >= sizes[path][idx1]:
                    bad_path = path
                    break
        return bad_path

    def _override_test(self, json_str, key, value, ovr_string,
                       host=None, group=None):
        error_msg = ''
        props = json.loads(json_str.strip())
        for prop in props:
            if group is not None:
                if prop['Group'] == group:
                    error_msg = self._check_override_value(prop, key,
                                                           value, ovr_string)
            elif host is not None:
                if prop['Host'] == host:
                    error_msg = self._check_override_value(prop, key,
                                                           value, ovr_string)
            else:
                error_msg = self._check_override_value(prop, key,
                                                       value, ovr_string)
        return error_msg

    def _check_override_value(self, prop, key, value, ovr_string):
        error_msg = ''
        if(prop['Property Name'] == key and
           prop['Property Value'] == value and
           prop['OVR'] != ovr_string):
            error_msg = ('override value mismatch for '
                         'key:%s value:%s ovr:%s target:%s' %
                         (key, value, prop['OVR'], ovr_string))
        return error_msg


if __name__ == '__main__':
    unittest.main()



# Copyright(c) 2016, Oracle and/or its affiliates.  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
import getpass
import os
import traceback

from cliff.command import Command
from cliff.lister import Lister

from kolla_cli.api.client import ClientApi
from kolla_cli.commands.exceptions import CommandError
import kolla_cli.i18n as u

CLIENT = ClientApi()


class PasswordSet(Command):
    "Password Set"

    def get_parser(self, prog_name):
        parser = super(PasswordSet, self).get_parser(prog_name)
        parser.add_argument('passwordname', metavar='<passwordname>',
                            help=u._('Password name'))
        parser.add_argument('--insecure', nargs='?', default=False,
                            help=argparse.SUPPRESS)
        return parser

    def take_action(self, parsed_args):
        try:
            password_name = parsed_args.passwordname.strip()
            if parsed_args.insecure is not False:
                # --insecure flag is present
                password = ''  # nosec
                if parsed_args.insecure:
                    password = parsed_args.insecure.strip()
            else:
                password = getpass.getpass(u._('Password: ')).strip()
                passtwo = getpass.getpass(u._('Retype Password: ')).strip()

                if password != passtwo:
                    raise CommandError(u._('Passwords do not match'))

            CLIENT.password_set(password_name, password)

        except Exception:
            raise Exception(traceback.format_exc())


class PasswordSetKey(Command):
    "Password Set SSH Key"

    def get_parser(self, prog_name):
        parser = super(PasswordSetKey, self).get_parser(prog_name)
        parser.add_argument('passwordname', metavar='<passwordname>',
                            help=u._('Password name'))
        parser.add_argument('privatekeypath', metavar='<privatekeypath>',
                            help=u._('Path to private key file'))
        parser.add_argument('publickeypath', metavar='<publickeypath>',
                            help=u._('Path to public key file'))
        return parser

    def take_action(self, parsed_args):
        try:
            password_name = parsed_args.passwordname.strip()
            private_keypath = parsed_args.privatekeypath.strip()
            private_keypath = os.path.abspath(private_keypath)
            public_keypath = parsed_args.publickeypath.strip()
            public_keypath = os.path.abspath(public_keypath)

            if not os.path.isfile(private_keypath):
                raise(CommandError(u._('Private key file not found: {path}')
                                   .format(path=private_keypath)))
            if not os.path.isfile(public_keypath):
                raise(CommandError(u._('Public key file not found: {path}')
                                   .format(path=public_keypath)))

            with open(private_keypath, 'r') as f:
                private_key = f.read()
            with open(public_keypath, 'r') as f:
                public_key = f.read()
            CLIENT.password_set_sshkey(password_name, private_key.strip(),
                                       public_key.strip())

        except Exception:
            raise Exception(traceback.format_exc())


class PasswordClear(Command):
    "Password Clear"

    def get_parser(self, prog_name):
        parser = super(PasswordClear, self).get_parser(prog_name)
        parser.add_argument('passwordname', metavar='<passwordname>',
                            help=u._('Password name'))
        return parser

    def take_action(self, parsed_args):
        try:
            password_name = parsed_args.passwordname.strip()
            CLIENT.password_clear(password_name)
        except Exception:
            raise Exception(traceback.format_exc())


class PasswordList(Lister):
    """List all password names."""

    def take_action(self, parsed_args):
        try:
            password_names = CLIENT.password_get_names()
            password_names = sorted(password_names)

            data = []
            for password_name in password_names:
                data.append((password_name, '-'))

            return ((u._('Password Name'),  u._('Password')), data)
        except Exception:
            raise Exception(traceback.format_exc())


class PasswordInit(Command):
    """Init all empty passwords and ssh keys."""

    def take_action(self, parsed_args):
        try:
            CLIENT.password_init()
        except Exception:
            raise Exception(traceback.format_exc())



# Copyright(c) 2017, Oracle and/or its affiliates.  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import fcntl
import grp
import logging
import os
import pwd
import subprocess  # nosec
import time
import yaml

from kolla_cli.api.exceptions import InvalidArgument
from kolla_cli.api.exceptions import MissingArgument
from kolla_cli.commands.exceptions import CommandError
import kolla_cli.i18n as u

LOG = logging.getLogger(__name__)

private_key_string = 'private_key'
public_key_string = 'public_key'


def get_log_level():
    evar = os.environ.get('KOLLA_LOG_LEVEL', 'info')
    if evar.lower() == 'debug':
        level = logging.DEBUG
    else:
        level = logging.INFO
    return level


def get_kolla_ansible_home():
    return os.environ.get("KOLLA_HOME", "/usr/share/kolla-ansible/")


def get_kolla_etc():
    return os.environ.get('KOLLA_ETC', '/etc/kolla/')


def get_kolla_cli_home():
    return os.path.join(get_kolla_ansible_home(), 'kolla-cli')


def get_kolla_cli_etc():
    return os.path.join(get_kolla_etc(), 'kolla-cli')


def get_group_vars_dir():
    return os.path.join(get_kolla_ansible_home(), 'ansible/group_vars')


def get_host_vars_dir():
    return os.path.join(get_kolla_ansible_home(), 'ansible/host_vars')


def get_tools_path():
    return os.environ.get(
        'KOLLA_TOOLS_DIR', os.path.join(get_kolla_cli_home(), 'tools'))


def get_kolla_actions_path():
    return os.path.join(get_tools_path(), 'kolla_actions.py')


def get_admin_uids():
    """get uid and gid of admin user"""
    user_info = pwd.getpwnam(get_admin_user())
    uid = user_info.pw_uid
    gid = user_info.pw_gid
    return uid, gid


def get_property_list_length():
    envvar = 'KOLLA_PROP_LIST_LENGTH'
    length_str = os.environ.get(envvar, '50')
    try:
        length = int(length_str)
    except Exception:
        raise InvalidArgument(
            u._('Environmental variable ({env_var}) is not an '
                'integer ({prop_length}).')
            .format(env_var=envvar, prop_length=length_str))
    return length


def get_admin_user():
    return os.environ.get("KOLLA_CLI_ADMIN_USER", "root")


def get_setup_user():
    return os.environ.get("KOLLA_CLI_SETUP_USER", "root")


def get_lock_enabled():
    evar = os.environ.get('KOLLA_CLI_LOCK', 'true')
    if evar.lower() == 'false':
        return False
    else:
        return True


def get_ansible_command(playbook=False):
    """Get the ansible command"""
    cmd = 'ansible'
    if playbook:
        cmd = 'ansible-playbook'
    return cmd


def run_cmd(cmd, print_output=True):
    """run a system command

    return:
        err_msg:
            empty string=command succeeded not None=command failed
        output:
            string: all the output of the run command
    """
    output = None
    try:
        process = subprocess.Popen(cmd, shell=True,  # nosec
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        output, err = process.communicate()
    except Exception as e:
        err = str(e)

    err = safe_decode(err)
    output = safe_decode(output)
    if process is not None and process.returncode != 0:
        err = (u._('Command failed. : {error}')
               .format(error=err))
    if print_output:
        LOG.info(output)
    return err, output


def change_password(file_path, pname, pvalue=None, public_key=None,
                    private_key=None, clear=False):
    """change password in passwords.yml file

    file_path:         path to passwords file
    pname:             name of password
    pvalue:            value of password when not ssh key
    public_key:        public ssh key
    private_key:       private ssh key
    clear:             flag to clear password

    If key is not found, an error is returned.
    If clear, and password exists, remove password.
    If clear, and password is already empty, nothing is done.
    If not clear, edit password in place.

    The passwords file contains both key-value pairs and key-dictionary
    pairs.  Type is maintained so you cannot change a key-dictionary
    password to a key-value password or the other way around.
    """
    read_data = sync_read_file(file_path)
    file_pwds = yaml.safe_load(read_data)
    # if the password file is empty file_pwds will be None after safe_load
    if file_pwds is None:
        file_pwds = {}

    if pname not in file_pwds.keys():
        raise Exception(
            u._('unable to update password as it does not exist: {pname}')
            .format(pname=pname))

    ssh_password_type = is_ssh_password(file_pwds[pname])

    if clear:
        # clear
        if pname in file_pwds:
            if ssh_password_type:
                file_pwds[pname] = {private_key_string: None,
                                    public_key_string: None}
            else:
                file_pwds[pname] = None
    else:
        # edit
        if private_key:
            if not ssh_password_type:
                raise Exception(
                    u._('unable to set non ssh type password to ssh value'))
            file_pwds[pname] = {private_key_string: private_key,
                                public_key_string: public_key}
        else:
            if ssh_password_type:
                raise Exception(
                    u._('unable to set ssh password type to non ssh value'))
            if not pvalue:
                pvalue = None
            file_pwds[pname] = pvalue

    # dump Nones as empty strings instead of the value 'null' as this is how
    # it looks when we read it.  also, this will not work with safe_dump
    yaml.add_representer(type(None), _empty_is_none)
    write_data = yaml.dump(file_pwds, default_flow_style=False)
    sync_write_file(file_path, write_data)


def clear_all_passwords():
    """clear all passwords in passwords.yml file"""
    password_path = os.path.join(get_kolla_etc(), 'passwords.yml')
    read_data = sync_read_file(password_path)
    file_pwds = yaml.safe_load(read_data)
    # if the password file is empty file_pwds will be None after safe_load
    if file_pwds is None:
        file_pwds = {}

    keys = file_pwds.keys()
    for key in keys:
        if is_ssh_password(file_pwds[key]):
            file_pwds[key] = {private_key_string: None,
                              public_key_string: None}
        else:
            file_pwds[key] = None

    yaml.add_representer(type(None), _empty_is_none)
    write_data = yaml.dump(file_pwds, default_flow_style=False)
    sync_write_file(password_path, write_data)


def _empty_is_none(self, _):
    return self.represent_scalar('tag:yaml.org,2002:null', '')


def is_ssh_password(password):
    if password is not None:
        if isinstance(password, dict):
            password_keys = password.keys()
            if (private_key_string in password_keys and
               public_key_string in password_keys):
                return True
    return False


def change_property(file_path, property_dict, clear=False):
    """change property with a file

    file_path:         path to property file
    property_dict:     dictionary of property names and values
    clear:             flag to remove property

    If clear, and property exists, remove it from the property file.
    If clear, and property doesn't exists, nothing is done.
    If not clear, and key is not found, the new property will be appended.
    If not clear, and key is found, edit property in place.
    """
    cloned_dict = copy.copy(property_dict)
    if not os.path.exists(file_path):
        with open(file_path, 'a'):
            os.utime(file_path, None)
            try:
                group_info = grp.getgrnam('kolla')
                os.chown(file_path, -1, group_info.gr_gid)
            except KeyError:
                # ignore error if kolla user not present, needed
                # for functional test
                LOG.debug('Ignoring error- kolla user not defined')

    new_contents = []
    read_data = sync_read_file(file_path)
    lines = read_data.split('\n')
    last_line_empty = False
    for line in lines:
        line = line.rstrip()

        # yank spurious empty lines
        if line:
            last_line_empty = False
        else:
            if last_line_empty:
                continue
            last_line_empty = True

        split_line = line.split(':', 1)
        if len(split_line) > 1:
            split_key = split_line[0]
            split_key.rstrip()
            if split_key in cloned_dict:
                if clear:
                    # clear existing property
                    continue
                # edit existing property
                value = cloned_dict[split_key]
                line = _get_property_line(split_key, value)

                # clear out the key after we are done, all existing keys
                # will be appended at the end (or for clear, ignored)
                del cloned_dict[split_key]
        new_contents.append(line)
    if not clear:
        # add new properties to file
        for key, value in cloned_dict.items():
            line = _get_property_line(key, value)

            # when we are doing an append we want to avoid
            # blank lines before the new entry
            if new_contents[-1:][0] == '':
                del new_contents[-1]
            new_contents.append(line)

    # if the last line is blank, trim it off
    if new_contents[-1:][0] == '':
        del new_contents[-1]
    write_data = '\n'.join(new_contents) + '\n'
    sync_write_file(file_path, write_data)


def _get_property_line(key, value):
    if type(value) is str:
        line = '%s: "%s"' % (key, value)
    else:
        str_value = yaml.safe_dump(value).strip()
        value_type = type(value)
        if value_type is bool or value_type is int:
            # yaml dump adds a newline and an ellipsis after
            # a boolean or int value. This needs to be stripped.
            str_value = str_value.replace('\n...', '')
        line = '%s: %s' % (key, str_value)
    return line


def sync_read_file(path, mode='r'):
    """synchronously read file

    return file data
    """
    lock = None
    try:
        if get_lock_enabled():
            lock = Lock(path, 'sync_read')
            locked = lock.wait_acquire()
            if not locked:
                raise Exception(
                    u._('unable to read file {path} '
                        'as it was locked.')
                    .format(path=path))
        with open(path, mode) as data_file:
            data = data_file.read()
    finally:
        if lock:
            lock.release()
    return safe_decode(data)


def sync_write_file(path, data, mode='w'):
    """synchronously write file"""
    ansible_lock = None
    lock = None
    try:
        if get_lock_enabled():
            ansible_lock = Lock(owner='sync_write')
            locked = ansible_lock.wait_acquire()
            if not locked:
                raise Exception(
                    u._('unable to get ansible lock while writing to {path} '
                        'as it was locked.')
                    .format(path=path))

            lock = Lock(lockpath=path, owner='sync_write')
            locked = lock.wait_acquire()
            if not locked:
                raise Exception(
                    u._('unable to write file {path} '
                        'as it was locked.')
                    .format(path=path))
        with open(path, mode) as data_file:
            data_file.write(data)
    finally:
        if ansible_lock:
            ansible_lock.release()
        if lock:
            lock.release()


def safe_decode(obj_to_decode):
    """Convert bytes or strings to unicode string

    Converts strings, lists, or dictionaries to
    unicode.
    """
    if obj_to_decode is None:
        return None

    if isinstance(obj_to_decode, list):
        new_obj = []
        for text in obj_to_decode:
            text = safe_decode(text)
            new_obj.append(text)
    elif isinstance(obj_to_decode, dict):
        new_obj = {}
        for key, value in obj_to_decode.items():
            key = safe_decode(key)
            value = safe_decode(value)
            new_obj[key] = value
    else:
        new_obj = obj_to_decode
        if not isinstance(obj_to_decode, str):
            # object is not unicode
            new_obj = obj_to_decode.decode('utf-8')
    return new_obj


def is_string_true(string):
    """Return boolean True if string represents a true value (None is False)"""
    true_values = ['yes', 'true']
    if string is not None and string.lower() in true_values:
        return True
    else:
        return False


def convert_lists_to_string(tuples, parsed_args):
    """convert lists to strings

    Because of the way cliff processes strings for tables, if a list
    has non-ascii chars in it, they would display as unicode bytes
    (\u0414\u0435\u043a\u0430\u0442). By converting
    the list to string here, the proper non-ascii chars are displayed.

    This will only change the lists when the output is to a user visible
    medium. It cannot be changed if the display output is json, yaml, etc.
    """
    convert_types = ['table', 'csv', 'html', 'value']
    if parsed_args.formatter and parsed_args.formatter not in convert_types:
        # not table output, leave it as-is
        return tuples

    new_tuples = []
    for data_tuple in tuples:
        new_items = []
        items = list(data_tuple)
        for item in items:
            if isinstance(item, list):
                item = convert_list_to_string(item)
            new_items.append(item)
        data_tuple = tuple(new_items)
        new_tuples.append(data_tuple)
    return new_tuples


def convert_list_to_string(alist):
    return '[' + ','.join(alist) + ']'


def check_arg(param, param_name, expected_type, none_ok=False, empty_ok=False,
              display_param=True):
    if param is None:
        if none_ok:
            return
        # None arg
        raise MissingArgument(param_name)

    if ((isinstance(param, str) or
            isinstance(param, dict) or
            isinstance(param, list)) and
            not param and not empty_ok):
        # empty string, dict or list
        raise MissingArgument(param_name)

    # if expected type is None, skip the type checking
    if expected_type is None:
        return

    if not isinstance(param, expected_type):
        # wrong type
        if display_param:
            raise InvalidArgument(u._('{name} ({param}) is not a {type}')
                                  .format(name=param_name, param=param,
                                          type=expected_type))
        else:
            raise InvalidArgument(u._('{name} is not a {type}')
                                  .format(name=param_name,
                                          type=expected_type))


def disallow_chars(param, param_name, chars):
    if param is None:
        return

    for char in chars:
        if char in param:
            raise InvalidArgument(
                u._('{name} contains invalid character {chars}')
                .format(name=param_name, chars=chars))


def handers_action_result(job, status, verbose_level):
    if verbose_level > 2:
        LOG.info('\n\n' + 80 * '=')
        LOG.info(u._('DEBUG command output:\n{out}')
                 .format(out=job.get_console_output()))
    if status == 0:
        if verbose_level > 1:
            # log any ansible warnings
            msg = job.get_error_message()
            if msg:
                LOG.warn(msg)
        LOG.info(u._('Success'))
    else:
        raise CommandError(u._('Job failed:\n{msg}')
                           .format(msg=job.get_error_message()))


class Lock(object):
    """Object which represents an exclusive resource lock

    flock usage is the default behavior but a separate pidfile mechanism
    is also available.  flock doesn't have the same orphaned lock issue
    that pidfile usage does.  both need to be tests on NFS.  if flock
    works then it seems better / less complicated for our needs.
    """

    def __init__(self, lockpath=None, owner='unknown owner', use_flock=True):
        self.lockpath = lockpath or self.get_lockpath()
        self.pid = str(os.getpid())
        self.fd = None
        self.owner = owner
        self.current_pid = -1
        self.current_owner = ''
        self.use_flock = use_flock

    def get_lockpath(self):
        return os.path.join(get_kolla_cli_home(), 'ansible.lock')

    def acquire(self):
        try:
            if self.use_flock:
                return self._acquire_flock()
            else:
                return self._acquire_pidfile()
        except IOError as e:
            # IOError is the error you get when the file is
            # already locked. (No such file returns an OSError.)
            # This may be OK and is handled by the caller.
            LOG.debug('Exception in acquiring lock. '
                      'path: %s pid: %s owner: %s error: %s' %
                      (self.lockpath, self.pid, self.owner, str(e)))
            return False
        except Exception as e:
            raise e

    def _acquire_pidfile(self):
        if not self.is_owned_by_me():
            fd = os.open(self.lockpath, os.O_CREAT | os.O_EXCL | os.O_RDWR)
            with os.fdopen(fd, 'a') as f:
                f.write(self.pid + '\n' + self.owner)
            return self.is_owned_by_me()

    def _acquire_flock(self):
        self.fd = os.open(self.lockpath, os.O_RDWR)
        fcntl.flock(self.fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
        return True

    def wait_acquire(self, wait_duration=3, interval=0.1):
        wait_time = 0
        while (wait_time < wait_duration):
            if not self.acquire():
                time.sleep(interval)
                wait_time += interval
            else:
                return True
        return False

    def is_owned_by_me(self):
        """Returns True if we own the lock or False otherwise"""
        try:
            if self.use_flock:
                raise Exception(u._('Invalid use of is_owned_by_me while'
                                    'using flock'))

            if not os.path.exists(self.lockpath):
                # lock doesn't exist, just return
                return False
            fd = os.open(self.lockpath, os.O_RDONLY)
            with os.fdopen(fd, 'r') as f:
                contents = f.read(2048).strip().split('\n')
                if len(contents) > 0:
                    self.current_pid = contents[0]
                if len(contents) > 1:
                    self.current_owner = contents[1]

                if contents[0] == str(self.pid):
                    return True
                else:
                    return False
        except Exception as e:
            # it is ok to fail to acquire, we just return that we failed
            LOG.debug('Exception in is_owned_by_me lock check. '
                      'path: %s pid: %s owner: %s error: %s' %
                      (self.lockpath, self.pid, self.owner, str(e)))
        return False

    def release(self):
        try:
            if self.use_flock:
                self._release_flock()
            else:
                self._release_pidfile()
        except OSError:
            # ignore release of an already released lock
            pass
        except Exception:
            # this really shouldn't happen unless for some reason
            # two areas in the same process try to release the lock
            # at the same time and if that happens you want to see
            # an error about it
            LOG.error('Error releasing lock', exc_info=True)
            return False

    def _release_pidfile(self):
        if self.is_owned_by_me():
            os.remove(self.lockpath)
            return True

    def _release_flock(self):
        try:
            fcntl.flock(self.fd, fcntl.LOCK_UN)
        except Exception as e:
            LOG.debug('Exception while releasing lock: %s' % str(e))
        finally:
            os.close(self.fd)
        return True


class PidManager(object):
    @staticmethod
    def get_child_pids(pid, child_pids=[]):
        """get child pids of parent pid"""
        # This ps command will return child pids of parent pid, separated by
        # newlines.
        err_msg, output = run_cmd('ps --ppid %s -o pid=""' % pid,
                                  print_output=False)

        # err_msg is expected when pid has no children
        if not err_msg:
            output = output.strip()

            if '\n' in output:
                ps_pids = output.split('\n')
            else:
                ps_pids = [output]

            if ps_pids:
                child_pids.extend(ps_pids)

                # recurse through children to get all child pids
                for ps_pid in ps_pids:
                    PidManager.get_child_pids(ps_pid, child_pids)
        return child_pids



# Copyright(c) 2016, Oracle and/or its affiliates.  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import os
import traceback

from kolla_cli.common.ansible.job import AnsibleJob
from kolla_cli.common.inventory import Inventory
from kolla_cli.common.utils import get_ansible_command
from kolla_cli.common.utils import get_kolla_etc

MYPY = False
if MYPY:
    from typing import List  # noqa


LOG = logging.getLogger(__name__)


class AnsiblePlaybook(object):
    playbook_path = ''
    extra_vars = ''
    include_passwords = True
    flush_cache = True
    print_output = True
    verbose_level = 0
    hosts = None  # type: List[str]
    groups = None  # type: List[str]
    services = None  # type: List[str]
    ignore_error_strings = None  # type: List[str]
    serial = False
    deploy_id = None  # type: str
    inventory = None  # type: Inventory
    local_only = False
    become_user = None  # type: str

    def run(self):
        try:
            if self.local_only:
                # Create a temporary local inventory with only localhost
                self.inventory = Inventory()
                self.inventory.set_deploy_mode(False)
                self.inventory.add_host('localhost')
            else:
                self.inventory = Inventory.load()
            inventory_path = self._make_temp_inventory()
            cmd = self._get_playbook_cmd(inventory_path)
            self._log_ansible_cmd(cmd, inventory_path)

            # create and run the job
            job = AnsibleJob(cmd, self.deploy_id,
                             self.print_output, inventory_path)
            job._ignore_error_strings = self.ignore_error_strings
            job.run()
            return job

        except Exception:
            raise Exception(traceback.format_exc())

    def _get_playbook_cmd(self, inventory_path):
        flag = ''
        # verbose levels: 1=not verbose, 2=more verbose
        if self.verbose_level > 1:
            flag = '-'
            for x in range(1, self.verbose_level):
                flag += 'v'

        ansible_cmd = get_ansible_command(playbook=True)
        cmd = '%s %s' % (ansible_cmd, flag)

        cmd += ' -i %s' % inventory_path

        if self.include_passwords:
            cmd += ' %s' % self._get_password_path()

        cmd += ' %s' % self.playbook_path

        if self.extra_vars or self.serial:
            extra_vars = ''
            if self.extra_vars:
                extra_vars = self.extra_vars
                if self.serial:
                    extra_vars += ' '
            if self.serial:
                extra_vars += 'serial_var=1'

            cmd += ' --extra-vars \"%s\"' % extra_vars

        if self.services:
            service_string = ''
            first = True
            for service in self.services:
                if not first:
                    service_string += ','
                else:
                    first = False
                service_string = service_string + service
            cmd += ' --tags %s' % service_string

        if self.hosts:
            host_string = ''
            first = True
            for host in self.hosts:
                if not first:
                    host_string += ','
                else:
                    first = False
                host_string = host_string + host
            cmd += ' --limit %s' % host_string

        if self.flush_cache:
            cmd += ' --flush-cache'

        if self.become_user:
            cmd += ' --become-user %s' % self.become_user

        return cmd

    def _make_temp_inventory(self):
        """Create temporary inventory file

        A temporary inventory is created so that a
        unique id can be assigned to the deployment.
        """
        inventory_filter = {}
        inventory_path = \
            self.inventory.create_json_gen_file(inventory_filter)

        # inv path = /tmp/kolla_UUID/temp_inventory.py
        deploy_id = os.path.dirname(inventory_path)
        self.deploy_id = deploy_id.split('kolla_')[1]

        return inventory_path

    def _get_password_path(self):
        kolla_etc = get_kolla_etc()
        return ('-e @' + os.path.join(kolla_etc, 'passwords.yml '))

    def _log_ansible_cmd(self, cmd, inventory_path):
        if self.verbose_level > 2:
            # log the ansible command
            LOG.debug('cmd:\n%s' % cmd)

            if self.verbose_level > 3:
                # log the inventory
                with open(inventory_path, 'r') as inv_file:
                    inv = inv_file.read()
                LOG.debug('\ninventory: \n%s' % inv)



# Copyright(c) 2018, Oracle and/or its affiliates.  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import traceback

from cliff.command import Command

from kolla_cli.api.client import ClientApi
from kolla_cli.commands.exceptions import CommandError
import kolla_cli.i18n as u

CLIENT = ClientApi()
LOG = logging.getLogger(__name__)


class ConfigReset(Command):
    """Resets the kolla-ansible configuration to its release defaults."""

    def take_action(self, parsed_args):
        try:
            CLIENT.config_reset()
        except Exception:
            raise Exception(traceback.format_exc())


class ConfigImport(Command):
    """Config Import

    """
    def get_parser(self, prog_name):
        parser = super(ConfigImport, self).get_parser(prog_name)
        parser.add_argument('import_type', metavar='<import_type>',
                            help=u._('Import type=<inventory>'))
        parser.add_argument('file_path', metavar='<file_path>',
                            help=u._('File path'))
        return parser

    def take_action(self, parsed_args):
        try:
            legal_types = ['inventory']
            import_type = parsed_args.import_type
            if not import_type or import_type not in legal_types:
                raise CommandError(u._(
                    'Import type must be {type}.').format(type=legal_types))

            file_path = None
            if parsed_args.file_path:
                file_path = parsed_args.file_path.strip()
            CLIENT.config_import_inventory(file_path)
        except Exception:
            raise Exception(traceback.format_exc())



# Copyright(c) 2018, Oracle and/or its affiliates.  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from kolla_cli.api.exceptions import FailedOperation
from kolla_cli.api.exceptions import InvalidArgument
from kolla_cli.common.inventory import Inventory
from kolla_cli.common import utils
from kolla_cli.common.utils import check_arg
import kolla_cli.i18n as u


class ConfigApi(object):

    @staticmethod
    def config_reset():
        """Config Reset.

        Resets the kolla-ansible configuration to its release defaults.
        """
        actions_path = utils.get_kolla_actions_path()
        cmd = ('%s config_reset' % actions_path)
        err_msg, output = utils.run_cmd(cmd, print_output=False)
        if err_msg:
            raise FailedOperation(
                u._('Configuration reset failed. {error} {message}')
                .format(error=err_msg, message=output))

    def config_import_inventory(self, file_path):
        # type: (str) -> None
        """Config Import Inventory

        Import groups and child associations from the provided
        inventory file. This currently does not import hosts, group
        vars, or host vars that may also exist in the inventory file.

        :param file_path: path to inventory file to import
        :type file_path: string
        """
        check_arg(file_path, u._('File path'), str)
        if not os.path.isfile(file_path):
            raise InvalidArgument(
                u._('File {path} is not valid.').format(
                    path=file_path))
        inventory = Inventory(file_path)
        Inventory.save(inventory)



# Copyright (c) 2018 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from kolla_cli.tests.unit.common import KollaCliUnitTest


class TestUnit(KollaCliUnitTest):
    @mock.patch('kolla_cli.api.client.ClientApi.group_add')
    @mock.patch('kolla_cli.shell.KollaCli._is_inventory_present',
                return_value=True)
    def test_group_add(self, _, mock_add):
        groupname = 'group1'
        ret = self.run_cli_command('group add %s' % groupname)
        self.assertEqual(ret, 0)
        mock_add.assert_called_once_with([groupname])

    @mock.patch('kolla_cli.api.client.ClientApi.group_remove')
    @mock.patch('kolla_cli.shell.KollaCli._is_inventory_present',
                return_value=True)
    def test_group_remove(self, _, mock_remove):
        groupname = 'group1'
        ret = self.run_cli_command('group remove %s' % groupname)
        self.assertEqual(ret, 0)
        mock_remove.assert_called_once_with([groupname])

    @mock.patch('kolla_cli.api.client.ClientApi.group_get_all')
    @mock.patch('kolla_cli.shell.KollaCli._is_inventory_present',
                return_value=True)
    def test_group_listhosts(self, _, mock_group_get_all):
        # list all groups and their hosts
        hostname = 'foo'
        groupname = 'group1'
        fake_group = self.get_fake_group(groupname, hostnames=[hostname])
        mock_group_get_all.return_value = [fake_group]
        ret = self.run_cli_command('group listhosts')
        self.assertEqual(ret, 0)
        mock_group_get_all.assert_called_once_with()

    @mock.patch('kolla_cli.api.group.Group.add_host')
    @mock.patch('kolla_cli.api.client.ClientApi.group_get')
    @mock.patch('kolla_cli.shell.KollaCli._is_inventory_present',
                return_value=True)
    def test_group_addhost(self, _, mock_group_get, mock_group_add_host):
        hostname = 'foo'
        groupname = 'group1'
        fake_group = self.get_fake_group(groupname)
        mock_group_get.return_value = [fake_group]
        ret = self.run_cli_command('group addhost %s %s'
                                   % (groupname, hostname))
        self.assertEqual(ret, 0)
        mock_group_get.assert_called_once_with([groupname])
        mock_group_add_host.assert_called_once_with(hostname)

    @mock.patch('kolla_cli.api.group.Group.remove_host')
    @mock.patch('kolla_cli.api.client.ClientApi.group_get')
    @mock.patch('kolla_cli.shell.KollaCli._is_inventory_present',
                return_value=True)
    def test_group_removehost(self, _, mock_group_get,
                              mock_group_remove_host):
        hostname = 'foo'
        groupname = 'group1'
        fake_group = self.get_fake_group(groupname, hostnames=[hostname])
        mock_group_get.return_value = [fake_group]
        ret = self.run_cli_command('group removehost %s %s'
                                   % (groupname, hostname))
        self.assertEqual(ret, 0)
        mock_group_get.assert_called_once_with([groupname])
        mock_group_remove_host.assert_called_once_with(hostname)

    @mock.patch('kolla_cli.api.client.ClientApi.group_get_all')
    @mock.patch('kolla_cli.shell.KollaCli._is_inventory_present',
                return_value=True)
    def test_group_listservices(self, _, mock_group_get_all):
        # list all groups and their services
        servicename = 'service1'
        groupname = 'group1'
        fake_group = self.get_fake_group(groupname,
                                         servicenames=[servicename])
        mock_group_get_all.return_value = [fake_group]
        ret = self.run_cli_command('group listservices')
        self.assertEqual(ret, 0)
        mock_group_get_all.assert_called_once_with()



# Copyright(c) 2016, Oracle and/or its affiliates.  All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#
from kolla_cli.tests.functional.common import KollaCliTest

import json
import unittest

from kolla_cli.api.client import ClientApi
from kolla_cli.api.exceptions import NotInInventory
from kolla_cli.common.ansible_inventory import AnsibleInventory

CLIENT = ClientApi()


class TestFunctional(KollaCliTest):

    def test_group_add_remove(self):
        groups = self.get_default_groups()

        # check default group list
        self.check_group(groups)

        tg1 = 'test_group_t1'
        tg2 = 'test_group_t2'

        self.run_cli_cmd('group add %s' % tg1)
        groups[tg1] = {
            'Services': [],
            'Hosts': []}
        self.check_group(groups)

        self.run_cli_cmd('group add %s' % tg2)
        groups[tg2] = {
            'Services': [],
            'Hosts': []}
        self.check_group(groups)

        self.run_cli_cmd('group remove %s' % tg2)
        del groups[tg2]
        self.check_group(groups)

        self.run_cli_cmd('group remove %s' % tg1)
        del groups[tg1]
        self.check_group(groups)

    def test_group_add_host(self):
        groups = self.get_default_groups()

        host1 = 'test_host1'
        host2 = 'test_host2'
        groupname = 'control'

        group = groups[groupname]
        hosts = group['Hosts']

        self.run_cli_cmd('host add %s' % host1)

        hosts.append(host1)
        self.run_cli_cmd('group addhost %s %s' % (groupname, host1))
        self.check_group(groups)

        self.run_cli_cmd('host add %s' % host2)

        hosts.append(host2)
        self.run_cli_cmd('group addhost %s %s' % (groupname, host2))
        self.check_group(groups)

        self.run_cli_cmd('group removehost %s %s' % (groupname, host2))
        hosts.remove(host2)
        self.check_group(groups)

        self.run_cli_cmd('group removehost %s %s' % (groupname, host1))
        hosts.remove(host1)
        self.check_group(groups)

    def test_add_group_to_service(self):
        groups = self.get_default_groups()

        groupname = 'network'
        service1 = 'keystone'
        service2 = 'heat-api'

        self.run_cli_cmd('service addgroup %s %s' % (service1, groupname))
        groups[groupname]['Services'].append(service1)
        self.check_group(groups)

        self.run_cli_cmd('service addgroup %s %s' % (service2, groupname))
        groups[groupname]['Services'].append(service2)
        self.check_group(groups)

        self.run_cli_cmd('service removegroup %s %s'
                         % (service2, groupname))
        groups[groupname]['Services'].remove(service2)
        self.check_group(groups)

        self.run_cli_cmd('service removegroup %s %s'
                         % (service1, groupname))
        groups[groupname]['Services'].remove(service1)
        self.check_group(groups)

    def test_group_api(self):
        # check some of the api not exercised by the CLI
        groupname1 = 'group_test1'
        groupname2 = 'group_test2'
        exp_groups = sorted([groupname1, groupname1, groupname2])
        CLIENT.group_add(exp_groups)
        groups = CLIENT.group_get([groupname1])
        groupnames = []
        for group in groups:
            groupnames.append(group.name)
        self.assertIn(groupname1, groupnames, 'group %s is missing in %s'
                      % (groupname1, groupnames))
        self.assertNotIn(groupname2, groupnames, 'group %s is unexpectedly in '
                         '%s' % (groupname2, groupnames))
        groups = CLIENT.group_get(exp_groups)
        groupnames = []
        for group in groups:
            groupnames.append(group.name)
        self.assertEqual(exp_groups, sorted(groupnames), 'groups mismatch')

        CLIENT.group_remove(exp_groups)
        try:
            CLIENT.group_get(exp_groups)
            self.assertTrue(False, 'Failed to raise NotInInventory exception')
        except NotInInventory:
            pass
        except Exception as e:
            raise e

        hostname1 = 'testhost1'
        CLIENT.group_add([groupname1])
        group1 = CLIENT.group_get([groupname1])[0]
        CLIENT.host_add([hostname1])
        group1.add_host(hostname1)
        hostnames = group1.get_hosts()
        self.assertIn(hostname1, hostnames, 'missing hostname')
        group1.add_service('nova')
        servicenames = group1.get_services()
        self.assertIn('nova', servicenames, 'missing servicename')

        # check the type checking logic
        self.check_types(CLIENT.group_add, [list])
        self.check_types(CLIENT.group_remove, [list])

    def check_group(self, groups):
        """check groups

        group listhosts -f json:
            [{"Group Name": "compute", "Hosts": []},
             {"Group Name": "control", "Hosts": ["ub-target1"]},
             {"Group Name": "network", "Hosts": []}]

        group listservices -f json:
            [{"Group Name": "compute", "Services": []},
             {"Group Name": "control",
              "Services": ["glance", "keystone", "mysqlcluster",
                           "nova", "rabbitmq"]},
             {"Group Name": "network",
              "Services": ["haproxy", "neutron"]}]
        """
        # check hosts in groups
        msg = self.run_cli_cmd('group listhosts -f json')
        cli_groups = json.loads(msg)
        self.assertEqual(len(cli_groups), len(groups),
                         '# of groups in cli not equal to expected groups.' +
                         '\n\nexpected: %s, \n\ncli: %s'
                         % (groups, cli_groups))

        for cli_group in cli_groups:
            cli_hosts = cli_group['Hosts']
            for group_name, info in groups.items():
                if group_name != cli_group['Group']:
                    continue
                group_hosts = info['Hosts']
                self.assertEqual(len(cli_hosts), len(group_hosts),
                                 'Group: %s. # of hosts in cli ' % group_name +
                                 'not equal to expected hosts, ' +
                                 '\n\nexpected: %s, \n\ncli: %s'
                                 % (group_hosts, cli_hosts))

                for group_host in group_hosts:
                    self.assertIn(group_host, cli_hosts,
                                  'Group: %s' % group_name +
                                  '\n\nexpected_hosts: %s, \n\nnot in cli: %s '
                                  % (group_host, cli_hosts))

        # check services in group
        msg = self.run_cli_cmd('group listservices -f json')
        cli_groups = json.loads(msg)

        for cli_group in cli_groups:
            cli_services = cli_group['Services']
            for group_name, info in groups.items():
                if group_name != cli_group['Group']:
                    continue
                group_services = info['Services']
                self.assertEqual(len(cli_services), len(group_services),
                                 'Group: %s. # of services in cli'
                                 % group_name +
                                 ' not equal to expected services,' +
                                 '\nexpected: %s, \ncli: %s'
                                 % (sorted(group_services),
                                    sorted(cli_services)))
                for group_service in group_services:
                    self.assertIn(group_service, cli_services,
                                  'Group: %s' % group_name +
                                  '\nexpected_services: %s, \nnot in cli: %s '
                                  % (sorted(group_services),
                                     sorted(cli_services)))

    def get_default_groups(self):
        """get default groups

        return a dict:
            {groupname: {
                 Services: [svc1, svc2...],
                 Hosts: []}}
        """
        ansible_inventory = AnsibleInventory()
        groupnames = ansible_inventory.groups
        groups = {}
        for groupname in groupnames:
            groups[groupname] = {'Services': [],
                                 'Hosts': []}

            for servicename, service in ansible_inventory.services.items():
                if groupname in service.get_groupnames():
                    groups[groupname]['Services'].append(servicename)

        return groups


if __name__ == '__main__':
    unittest.main()



# Copyright(c) 2016, Oracle and/or its affiliates.  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import traceback

from cliff.command import Command
from cliff.lister import Lister

from kolla_cli.api.client import ClientApi
from kolla_cli.api.exceptions import ClientException
from kolla_cli.commands.exceptions import CommandError
from kolla_cli.common.utils import convert_lists_to_string
import kolla_cli.i18n as u

CLIENT = ClientApi()


class GroupAdd(Command):
    """Add group to openstack-kolla."""
    def get_parser(self, prog_name):
        parser = super(GroupAdd, self).get_parser(prog_name)
        parser.add_argument('groupname', metavar='<groupname>',
                            help=u._('Group name'))
        return parser

    def take_action(self, parsed_args):
        try:
            groupname = parsed_args.groupname.strip()

            CLIENT.group_add([groupname])
        except ClientException as e:
            raise CommandError(str(e))
        except Exception:
            raise Exception(traceback.format_exc())


class GroupRemove(Command):
    """Remove group from openstack-kolla."""

    def get_parser(self, prog_name):
        parser = super(GroupRemove, self).get_parser(prog_name)
        parser.add_argument('groupname', metavar='<groupname>',
                            help=u._('Group name'))
        return parser

    def take_action(self, parsed_args):
        try:
            groupname = parsed_args.groupname.strip()

            CLIENT.group_remove([groupname])
        except ClientException as e:
            raise CommandError(str(e))
        except Exception:
            raise Exception(traceback.format_exc())


class GroupAddhost(Command):
    """Add host to group."""
    def get_parser(self, prog_name):
        parser = super(GroupAddhost, self).get_parser(prog_name)
        parser.add_argument('groupname', metavar='<groupname>',
                            help=u._('Group name'))
        parser.add_argument('hostname', metavar='<hostname>',
                            help=u._('Host name'))
        return parser

    def take_action(self, parsed_args):
        try:
            groupname = parsed_args.groupname.strip()
            hostname = parsed_args.hostname.strip()

            group = CLIENT.group_get([groupname])[0]
            group.add_host(hostname)

        except ClientException as e:
            raise CommandError(str(e))
        except Exception:
            raise Exception(traceback.format_exc())


class GroupRemovehost(Command):
    """Remove host group from group."""

    def get_parser(self, prog_name):
        parser = super(GroupRemovehost, self).get_parser(prog_name)
        parser.add_argument('groupname', metavar='<groupname>',
                            help=u._('Group name'))
        parser.add_argument('hostname', metavar='<hostname>',
                            help=u._('Host name'))
        return parser

    def take_action(self, parsed_args):
        try:
            groupname = parsed_args.groupname.strip()
            hostname = parsed_args.hostname.strip()

            group = CLIENT.group_get([groupname])[0]
            group.remove_host(hostname)

        except ClientException as e:
            raise CommandError(str(e))
        except Exception:
            raise Exception(traceback.format_exc())


class GroupList(Lister):
    """Only list all groups """

    def take_action(self, parsed_args):
        try:
            data = [('',)]
            groups = CLIENT.group_get_all()
            if groups:
                data = []
                for group in groups:
                    data.append((group.get_name(),))
            data = convert_lists_to_string(data, parsed_args)
            return ((u._('Group'), ), sorted(data))
        except ClientException as e:
            raise CommandError(str(e))
        except Exception:
            raise Exception(traceback.format_exc())


class GroupListhosts(Lister):
    """List all groups and their hosts."""

    def take_action(self, parsed_args):
        try:
            data = [('', '')]
            groups = CLIENT.group_get_all()
            if groups:
                data = []
                for group in groups:
                    data.append((group.get_name(),
                                 sorted(group.get_hosts())))
            data = convert_lists_to_string(data, parsed_args)
            return ((u._('Group'), u._('Hosts')), sorted(data))
        except ClientException as e:
            raise CommandError(str(e))
        except Exception:
            raise Exception(traceback.format_exc())


class GroupListservices(Lister):
    """List all groups and their services."""

    def take_action(self, parsed_args):
        try:
            data = [('', '')]
            groups = CLIENT.group_get_all()
            if groups:
                data = []
                for group in groups:
                    data.append((group.get_name(),
                                 sorted(group.get_services())))
            data = convert_lists_to_string(data, parsed_args)
            return ((u._('Group'), u._('Services')), sorted(data))
        except ClientException as e:
            raise CommandError(str(e))
        except Exception:
            raise Exception(traceback.format_exc())



# Copyright(c) 2016, Oracle and/or its affiliates.  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Command-line interface to Kolla"""
import logging
import os
import sys

from cliff.app import App
from cliff.commandmanager import CommandManager

import kolla_cli.i18n as u

from kolla_cli.api.client import ClientApi
from kolla_cli.commands.exceptions import CommandError
from kolla_cli.common.inventory import INVENTORY_PATH
from kolla_cli.common.utils import get_kolla_cli_etc

LOG = logging.getLogger(__name__)

VERSION = '4.0'


class KollaCli(App):
    def __init__(self):
        super(KollaCli, self).__init__(
            description=u._('Command-Line Client for OpenStack Kolla'),
            version=VERSION,
            command_manager=CommandManager('kolla.cli'),
            )

        inventory_path = os.path.join(get_kolla_cli_etc(),
                                      INVENTORY_PATH)
        if not self._is_inventory_present(inventory_path):
            err_string = u._(
                'Required file ({inventory}) does not exist.\n'
                'Please re-install the kollacli to '
                'recreate the file.').format(inventory=inventory_path)
            raise CommandError(err_string)

        # set up logging and test that user running shell is part
        # of kolla group
        ClientApi()

        # paramiko log is very chatty, tune it down
        logging.getLogger('paramiko').setLevel(logging.WARNING)

        self.dump_stack_trace = False

    def _is_inventory_present(self, inventory_path):
        return os.path.isfile(inventory_path)


def main(argv=sys.argv[1:]):
    shell = KollaCli()
    return shell.run(argv)


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



# Copyright(c) 2016, Oracle and/or its affiliates.  All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#
from kolla_cli.tests.functional.common import KollaCliTest

from kolla_cli.api.client import ClientApi

import unittest

TEST_GROUP_NAME = 'test_group'
CLIENT = ClientApi()

NOT_KNOWN = 'Name or service not known'


class TestFunctional(KollaCliTest):

    def test_stop(self):
        # No physical hosts in config, use a non-existent host.
        # This will generate expected exceptions in all host access
        # commands.
        hostnames = ['test_deploy_host1']
        CLIENT.host_add(hostnames)

        # add host to a new deploy group
        CLIENT.group_add([TEST_GROUP_NAME])
        group = CLIENT.group_get([TEST_GROUP_NAME])[0]
        for hostname in hostnames:
            group.add_host(hostname)

        # stop services, initialize server
        self.log.info('Start stop #1')
        job = CLIENT.stop(1, hostnames)
        self._process_job(job, 'stop #1')

        self.log.info('updating various properties for the test')

        # disable services so the test is quicker
        enable_service_props = {}
        for service in CLIENT.service_get_all():
            service_name = service.name.replace('-', '_')
            enable_service_props['enable_%s' % service_name] = 'no'
        CLIENT.property_set(enable_service_props)

        # do a deploy of a limited set of services
        self.log.info('Start a deployment')
        job = CLIENT.deploy()
        self._process_job(job, 'deploy')

        self.log.info('Start stop #2')
        job = CLIENT.stop(1, hostnames)
        self._process_job(job, 'stop #2')

    def _process_job(self, job, descr, expect_kill=False):
        status = job.wait()
        output = job.get_console_output()
        self.log.info('job is complete. status: %s, err: %s'
                      % (status, output))
        if expect_kill:
            self.assertEqual(2, status, 'Job %s does not have killed status %s'
                             % (descr, output))
        else:
            self.assertEqual(1, status, 'Job %s ' % descr +
                             'succeeded when it should have failed')


if __name__ == '__main__':
    unittest.main()



# Copyright(c) 2016, Oracle and/or its affiliates.  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
import getpass
import logging
import os
import traceback
import yaml

from cliff.command import Command
from cliff.lister import Lister
from six.moves import input

from kolla_cli.api.client import ClientApi
from kolla_cli.api.exceptions import ClientException
from kolla_cli.commands.exceptions import CommandError
from kolla_cli.common.utils import convert_lists_to_string
from kolla_cli.common.utils import get_setup_user
from kolla_cli.common.utils import handers_action_result
import kolla_cli.i18n as u

LOG = logging.getLogger(__name__)
CLIENT = ClientApi()


class HostAdd(Command):
    """Add host to openstack-kolla."""

    def get_parser(self, prog_name):
        parser = super(HostAdd, self).get_parser(prog_name)
        parser.add_argument('hostname', metavar='<hostname>',
                            help=u._('Host name or ip address'))
        return parser

    def take_action(self, parsed_args):
        try:
            hostname = parsed_args.hostname.strip()
            CLIENT.host_add([hostname])

        except ClientException as e:
            raise CommandError(str(e))
        except Exception:
            raise Exception(traceback.format_exc())


class HostDestroy(Command):
    """Destroy all kolla containers on host(s).

    Stops and removes all kolla related docker containers on either the
    specified host or all hosts if the hostname all is used.
    """

    def get_parser(self, prog_name):
        parser = super(HostDestroy, self).get_parser(prog_name)
        parser.add_argument('hostname', metavar='<hostname | all>',
                            help=u._('Host name or ip address or "all"'))
        parser.add_argument('--stop', action='store_true',
                            help=u._('Stop rather than kill'))
        parser.add_argument('--includedata', action='store_true',
                            help=u._('Destroy data containers'))
        parser.add_argument('--removeimages', action='store_true',
                            help=u._('Remove docker images'))
        return parser

    def take_action(self, parsed_args):
        try:
            hostname = parsed_args.hostname.strip()

            hostnames = [hostname]
            if hostname == 'all':
                hostnames = _get_all_hostnames()
                # if there are no hosts, don't bother doing anything
                if not hostnames:
                    return

            destroy_type = 'kill'
            if parsed_args.stop:
                destroy_type = 'stop'
            include_data = False
            if parsed_args.includedata:
                include_data = True
            remove_images = False
            if parsed_args.removeimages:
                remove_images = True

            if include_data and not self._is_ok_to_delete_data():
                LOG.info('Aborting destroy')
                return

            verbose_level = self.app.options.verbose_level

            job = CLIENT.host_destroy(hostnames, destroy_type,
                                      verbose_level, include_data,
                                      remove_images)
            status = job.wait()
            handers_action_result(job, status, verbose_level)
        except ClientException as e:
            raise CommandError(str(e))
        except Exception:
            raise Exception(traceback.format_exc())

    def _is_ok_to_delete_data(self):
        question = ('This will delete all containers and data'
                    ', are you sure? (y/n)')
        answer = input(question)
        while answer != 'y' and answer != 'n':
            answer = input(question)
        return True if answer == 'y' else False


class HostRemove(Command):
    """Remove host from openstack-kolla."""

    def get_parser(self, prog_name):
        parser = super(HostRemove, self).get_parser(prog_name)
        parser.add_argument('hostname', metavar='<hostname | all>',
                            help=u._('Host name or "all"'))
        return parser

    def take_action(self, parsed_args):
        try:
            hostname = parsed_args.hostname.strip()
            hostnames = [hostname]
            if hostname == 'all':
                hostnames = _get_all_hostnames()
            if len(hostnames) > 0:
                CLIENT.host_remove(hostnames)

        except ClientException as e:
            raise CommandError(str(e))
        except Exception:
            raise Exception(traceback.format_exc())


class HostList(Lister):
    """List hosts and their groups.

    If a hostname is provided, only list information about that host.
    """

    def get_parser(self, prog_name):
        parser = super(HostList, self).get_parser(prog_name)
        parser.add_argument('hostname', nargs='?', metavar='[hostname]',
                            help=u._('Host name'))
        return parser

    def take_action(self, parsed_args):
        try:
            hostname = None
            if parsed_args.hostname:
                hostname = parsed_args.hostname.strip()

            hosts = []
            if hostname:
                hosts = CLIENT.host_get([hostname])
            else:
                hosts = CLIENT.host_get_all()

            data = []
            if hosts:
                for host in hosts:
                    data.append((host.name, host.get_groups()))
            else:
                data.append(('', ''))

            data = convert_lists_to_string(data, parsed_args)
            return ((u._('Host'), u._('Groups')), sorted(data))

        except ClientException as e:
            raise CommandError(str(e))
        except Exception:
            raise Exception(traceback.format_exc())


class HostCheck(Command):
    """Check an ssh check of host(s)."""

    def get_parser(self, prog_name):
        parser = super(HostCheck, self).get_parser(prog_name)
        parser.add_argument('hostname', metavar='<hostname | all>',
                            help=u._('Host name or "all"'))
        return parser

    def take_action(self, parsed_args):
        try:
            hostname = parsed_args.hostname.strip()
            if hostname == 'all':
                hostnames = _get_all_hostnames()
            else:
                hostnames = [hostname]

            # just do an ssh check
            summary = CLIENT.host_ssh_check(hostnames)
            all_ok = True
            for hostname, info in summary.items():
                status = u._('success')
                msg = ''
                if not info['success']:
                    status = u._('failed-')
                    msg = info['msg']
                    all_ok = False
                LOG.info(u._('Host {host}: {sts} {msg}')
                         .format(host=hostname, sts=status, msg=msg))

            if not all_ok:
                raise CommandError(u._('Host check failed.'))
        except ClientException as e:
            raise CommandError(str(e))
        except Exception:
            raise Exception(traceback.format_exc())


class HostSetup(Command):
    """Setup kolla-cli on host."""

    def get_parser(self, prog_name):
        parser = super(HostSetup, self).get_parser(prog_name)
        parser.add_argument('hostname', nargs='?',
                            metavar='<hostname>', help=u._('Host name'))
        parser.add_argument('--insecure', nargs='?', help=argparse.SUPPRESS)
        parser.add_argument('--file', '-f', nargs='?',
                            metavar='<hosts_info_file>',
                            help=u._('Absolute path to hosts info file '))
        return parser

    def take_action(self, parsed_args):
        try:
            if not parsed_args.hostname and not parsed_args.file:
                raise CommandError(
                    u._('Host name or hosts info file path is required.'))
            if parsed_args.hostname and parsed_args.file:
                raise CommandError(
                    u._('Host name and hosts info file path '
                        'cannot both be present.'))

            if parsed_args.file:
                # multi-host setup via xml file
                # The xml file's content is like a dict of format:
                # hostname1:
                #     uname: user1
                #     password: <password1>
                # hostname2:
                #     uname: user2
                #     password: <password2>
                hosts_data = self._get_yml_data(parsed_args.file.strip())
                CLIENT.host_setup(hosts_data)
            else:
                # single host setup
                hostname = parsed_args.hostname.strip()
                summary = CLIENT.host_ssh_check([hostname])
                if summary[hostname]['success']:
                    LOG.info(
                        u._LI('Skipping setup of host ({host}) as '
                              'ssh check is ok.').format(host=hostname))
                    return 0

                if parsed_args.insecure:
                    password = parsed_args.insecure.strip()
                else:
                    password = getpass.getpass(
                        u._('{name} password for {host}: ')
                        .format(name=get_setup_user(), host=hostname))
                CLIENT.host_setup({hostname: {'password': password}})

        except ClientException as e:
            raise CommandError(str(e))
        except Exception:
            raise Exception(traceback.format_exc())

    def _get_yml_data(self, yml_path):
        if not os.path.isfile(yml_path):
            raise CommandError(
                u._('No file exists at {path}. An absolute file path is '
                    'required.').format(path=yml_path))

        with open(yml_path, 'r') as hosts_file:
            file_data = hosts_file.read()

        hosts_info = yaml.safe_load(file_data)
        if not hosts_info:
            raise CommandError(u._('{path} is empty.').format(path=yml_path))
        return hosts_info


def _get_all_hostnames():
    hostnames = []
    hosts = CLIENT.host_get_all()
    for host in hosts:
        hostnames.append(host.name)
    return hostnames



# Copyright(c) 2016, Oracle and/or its affiliates.  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import logging
import os
import yaml

from kolla_cli.api.exceptions import NotInInventory
from kolla_cli.common.inventory import Inventory
from kolla_cli.common.utils import change_property
from kolla_cli.common.utils import get_group_vars_dir
from kolla_cli.common.utils import get_host_vars_dir
from kolla_cli.common.utils import get_kolla_ansible_home
from kolla_cli.common.utils import sync_read_file
import kolla_cli.i18n as u

LOG = logging.getLogger(__name__)

ALLVARS_PATH = 'ansible/group_vars/all.yml'
GLOBALS_PATH = 'ansible/group_vars/__GLOBAL__'
ANSIBLE_ROLES_PATH = 'ansible/roles'
ANSIBLE_DEFAULTS_PATH = 'defaults/main.yml'


class AnsibleProperties(object):

    def __init__(self):
        """initialize ansible property information

        property information is pulled from the following files
        (from lowest to highest priority):
        KOLLA_HOME/ansible/roles/<service>/default/main.yml
        KOLLA_HOME/ansible/group_vars/all.yml
        KOLLA_HOME/ansible/group_vars/__GLOBAL__
        KOLLA_HOME/ansible/group_vars/*
        KOLLA_HOME/ansible/host_vars/*
        KOLLA_ETC/passwords.yml
        """
        self.globals_path = os.path.join(get_kolla_ansible_home(),
                                         GLOBALS_PATH)
        self.global_props = []
        self.unique_global_props = {}
        self.unique_override_flags = {}
        self.group_props = {}
        self.host_props = {}
        self.properties_loaded = False
        self._inventory = None

    def _load_properties(self):
        self._load_inventory()
        if not self.properties_loaded:
            self._load_properties_roles()
            self._load_properties_all()
            self._load_properties_global()
            self._load_properties_hostvars()
            self._load_properties_groupvars()
            self.properties_loaded = True

    def _load_properties_roles(self):
        start_dir = os.path.join(get_kolla_ansible_home(), ANSIBLE_ROLES_PATH)
        services = next(os.walk(start_dir))[1]
        for service_name in services:
            file_name = os.path.join(start_dir, service_name,
                                     ANSIBLE_DEFAULTS_PATH)
            if os.path.isfile(file_name):
                with open(file_name) as service_file:
                    service_contents = yaml.safe_load(service_file)
                    prop_file_name = service_name + ':main.yml'
                    for key, value in service_contents.items():
                        ansible_prop = AnsibleProperty(key, value,
                                                       prop_file_name)
                        self.global_props.append(ansible_prop)
                        self.unique_global_props[key] = ansible_prop

    def _load_properties_all(self):
        allvars_path = os.path.join(get_kolla_ansible_home(), ALLVARS_PATH)
        with open(allvars_path) as allvars_file:
            allvars_contents = yaml.safe_load(allvars_file)
            for key, value in allvars_contents.items():
                overrides = False
                orig_value = None
                if key in self.unique_global_props:
                    overrides = True
                    orig_value = self.unique_global_props[key].value
                ansible_prop = AnsibleProperty(key, value,
                                               'group_vars/all.yml',
                                               overrides, orig_value)
                self.global_props.append(ansible_prop)
                self.unique_global_props[key] = ansible_prop

    def _load_properties_global(self):
        globals_data = sync_read_file(self.globals_path)
        globals_contents = yaml.safe_load(globals_data)
        if not globals_contents:
            return
        for key, value in globals_contents.items():
            overrides = False
            override_flags = OverrideFlags()
            orig_value = None
            if key in self.unique_global_props:
                overrides = True
                override_flags.ovr_global = True
                orig_value = self.unique_global_props[key].value
            ansible_prop = AnsibleProperty(key, value,
                                           'group_vars/__GLOBAL__',
                                           overrides, orig_value)
            ansible_prop.override_flags = override_flags
            self.global_props.append(ansible_prop)
            self.unique_global_props[key] = ansible_prop
            self.unique_override_flags[key] = override_flags

    def _load_properties_hostvars(self):
        host_dir = get_host_vars_dir()
        hostnames = self._inventory.get_hostnames()
        for hostfile in os.listdir(host_dir):
            if hostfile not in hostnames:
                # skip any host files that don't match existing hosts
                continue
            self.host_props[hostfile] = []
            with open(os.path.join(host_dir, hostfile)) as host_data:
                host_contents = yaml.safe_load(host_data)
                if host_contents is None:
                    continue
                props = []
                for key, value in host_contents.items():
                    overrides = False
                    override_flags = OverrideFlags()
                    if key in self.unique_override_flags:
                        override_flags = self.unique_override_flags[key]
                    orig_value = None
                    if key in self.unique_global_props:
                        overrides = True
                        override_flags.ovr_host = True
                        self.unique_override_flags[key] = override_flags
                        orig_value = self.unique_global_props[key].value
                    ansible_prop = AnsibleProperty(key, value,
                                                   hostfile,
                                                   overrides, orig_value,
                                                   'host', hostfile)
                    props.append(ansible_prop)
            self.host_props[hostfile] = props

    def _load_properties_groupvars(self):
        group_dir = get_group_vars_dir()
        groupnames = self._inventory.get_groupnames()
        for groupfile in os.listdir(group_dir):
            if groupfile not in groupnames:
                # skip any files that don't match existing groups
                continue
            with open(os.path.join(group_dir, groupfile)) as group_data:
                group_contents = yaml.safe_load(group_data)
                if group_contents is None:
                    continue
                props = []
                for key, value in group_contents.items():
                    overrides = False
                    override_flags = OverrideFlags()
                    if key in self.unique_override_flags:
                        override_flags = self.unique_override_flags[key]
                    orig_value = None
                    if key in self.unique_global_props:
                        overrides = True
                        override_flags.ovr_group = True
                        self.unique_override_flags[key] = override_flags
                        orig_value = self.unique_global_props[key].value
                    ansible_prop = AnsibleProperty(key, value,
                                                   groupfile,
                                                   overrides, orig_value,
                                                   'group', groupfile)
                    props.append(ansible_prop)
            self.group_props[groupfile] = props

    def _load_inventory(self):
        if not self._inventory:
            self._inventory = Inventory.load()  # nosec

    def get_host_list(self, host_list):
        self._load_properties()
        prop_list = []
        if host_list is not None:
            for host_name in host_list:
                host = self._inventory.get_host(host_name)
                if host is None:
                    raise NotInInventory(u._('Host'), host_name)
                if host_name in self.host_props:
                    prop_list += self.host_props[host_name]
        else:
            hosts = self._inventory.get_hosts()
            for host in hosts:
                if host.name in self.host_props:
                    prop_list += self.host_props[host.name]
        return prop_list

    def get_group_list(self, group_list):
        self._load_properties()
        prop_list = []
        if group_list is not None:
            for group_name in group_list:
                group = self._inventory.get_group(group_name)
                if group is None:
                    raise NotInInventory(u._('Group'), group_name)
                if group_name in self.group_props:
                    prop_list += self.group_props[group_name]
        else:
            groups = self._inventory.get_groups()
            for group in groups:
                if group.name in self.group_props:
                    prop_list += self.group_props[group.name]
        return prop_list

    def get_property_value(self, property_name):
        self._load_properties()
        prop_val = None
        if property_name in self.unique_global_props:
            prop = self.unique_global_props[property_name]
            prop_val = prop.value
        return prop_val

    def get_property(self, property_name):
        self._load_properties()
        return self.unique_global_props.get(property_name)

    def get_all_unique(self):
        self._load_properties()
        unique_list = []
        for _, value in self.unique_global_props.items():
            unique_list.append(value)
        return sorted(unique_list, key=lambda x: x.name)

    def get_all_override_flags(self):
        self._load_properties()
        return self.unique_override_flags

    def set_property(self, property_dict):
        change_property(self.globals_path, property_dict,
                        clear=False)

    def set_host_property(self, property_dict, hosts):
        # if hosts is None set the property on all hosts
        self._load_inventory()
        host_list = []
        if hosts is None:
            host_list = self._inventory.get_hosts()
        else:
            for host_name in hosts:
                host = self._inventory.get_host(host_name)
                if host is None:
                    raise NotInInventory(u._('Host'), host_name)
                host_list.append(host)
        try:
            for host in host_list:
                file_path = os.path.join(get_host_vars_dir(), host.name)
                change_property(file_path, property_dict,
                                clear=False)
        except Exception as e:
            raise e

    def set_group_property(self, property_dict, groups):
        # if groups is None set the property on all hosts
        self._load_inventory()
        group_list = []
        if groups is None:
            group_list = self._inventory.get_groups()
        else:
            for group_name in groups:
                group = self._inventory.get_group(group_name)
                if group is None:
                    raise NotInInventory(u._('Group'), group_name)
                group_list.append(group)
        try:
            for group in group_list:
                tmp_dict = copy.copy(property_dict)
                file_path = os.path.join(get_group_vars_dir(), group.name)
                change_property(file_path, tmp_dict,
                                clear=False)
        except Exception as e:
            raise e

    def clear_property(self, property_list):
        try:
            change_property(self.globals_path,
                            self._list_to_dict(property_list),
                            clear=True)
        except Exception as e:
            raise e

    def clear_host_property(self, property_list, hosts):
        # if hosts is None set the property on all hosts
        self._load_inventory()
        host_list = []
        if hosts is None:
            host_list = self._inventory.get_hosts()
        else:
            for host_name in hosts:
                host = self._inventory.get_host(host_name)
                if host is None:
                    raise NotInInventory(u._('Host'), host_name)
                host_list.append(host)
        try:
            for host in host_list:
                file_path = os.path.join(get_host_vars_dir(), host.name)
                change_property(file_path, self._list_to_dict(property_list),
                                clear=True)
        except Exception as e:
            raise e

    def clear_group_property(self, property_list, groups):
        # if hosts is None set the property on all hosts
        self._load_inventory()
        group_list = []
        if groups is None:
            group_list = self._inventory.get_groups()
        else:
            for group_name in groups:
                group = self._inventory.get_group(group_name)
                if group is None:
                    raise NotInInventory(u._('Group'), group_name)
                group_list.append(group)
        try:
            for group in group_list:
                file_path = os.path.join(get_group_vars_dir(), group.name)
                change_property(file_path, self._list_to_dict(property_list),
                                clear=True)
        except Exception as e:
            raise e

    def _list_to_dict(self, property_list):
        property_dict = {}
        for key in property_list:
            property_dict[key] = ''
        return property_dict


class AnsibleProperty(object):

    def __init__(self, name, value, file_name, overrides=False,
                 orig_value=None, prop_type='global', target=None):
        self.name = name
        self.value = value
        self.prop_type = prop_type
        self.file_name = file_name
        self.overrides = overrides
        self.orig_value = orig_value
        self.target = target
        self.value_type = type(value)


class OverrideFlags(object):

    def __init__(self):
        self.ovr_global = False
        self.ovr_group = False
        self.ovr_host = False



# Copyright(c) 2016, Oracle and/or its affiliates.  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import os

from kolla_cli.api.exceptions import InvalidArgument
from kolla_cli.api.exceptions import InvalidConfiguration
from kolla_cli.api.exceptions import NotInInventory
from kolla_cli.common.ansible.playbook import AnsiblePlaybook
from kolla_cli.common.inventory import Inventory
from kolla_cli.common.passwords import get_empty_password_values
from kolla_cli.common.properties import AnsibleProperties
from kolla_cli.common.utils import get_admin_user
from kolla_cli.common.utils import get_kolla_ansible_home
from kolla_cli.common.utils import get_kolla_etc
from kolla_cli.common.utils import is_string_true
import kolla_cli.i18n as u

LOG = logging.getLogger(__name__)


class KollaAction(object):
    """Kolla Action."""

    def __init__(self, verbose_level=0, playbook_name=''):
        self.playbook_name = playbook_name
        self.playbook_path = os.path.join(get_kolla_ansible_home(),
                                          'ansible/',
                                          self.playbook_name)
        self.playbook = AnsiblePlaybook()
        self.playbook.verbose_level = verbose_level
        self.playbook.playbook_path = self.playbook_path

    def certificate_init(self):
        '''Creates a self-signed certificate'''

        self.playbook.local_only = True
        self.playbook.become_user = get_admin_user()
        job = self.playbook.run()
        return job

    def deploy(self, hostnames=[], serial_flag=False, servicenames=[]):
        '''Deploy and start all kolla containers.'''

        self.playbook.hosts = hostnames
        self.playbook.serial = serial_flag
        self.playbook.services = servicenames
        self.playbook.extra_vars = 'kolla_action=deploy'

        self._run_deploy_rules(self.playbook)

        job = self.playbook.run()
        return job

    def reconfigure(self, hostnames=[], servicenames=[]):
        '''Reconfigure OpenStack service.'''

        self.playbook.hosts = hostnames
        self.playbook.services = servicenames
        self.playbook.extra_vars = 'kolla_action=reconfigure'

        self._run_deploy_rules(self.playbook)

        job = self.playbook.run()
        return job

    def postdeploy(self):
        '''Do post deploy on deploy node.'''

        self.playbook.local_only = True
        self.playbook.become_user = get_admin_user()
        job = self.playbook.run()
        return job

    def destroy_hosts(self, hostnames, destroy_type,
                      include_data=False, remove_images=False):
        '''destroy containers on a set of hosts.

        The containers on the specified hosts will be stopped
        or killed.
        '''

        LOG.info(u._LI('Please be patient as this may take a while.'))
        # 'hosts' is defined as 'all' in the playbook yml code, but inventory
        # filtering will subset that down to the hosts in playbook.hosts.
        self.playbook.hosts = hostnames
        if remove_images:
            self.playbook.extra_vars = 'destroy_include_images=yes'
        if self.playbook.verbose_level <= 1:
            self.playbook.print_output = False
        job = self.playbook.run()
        return job

    def stop(self, hostnames=[], servicenames=[]):
        '''stop containers on a set of hosts.

        The containers on the specified hosts will be stopped
        or killed if the stop takes over 20 seconds.
        '''

        LOG.info(u._LI('Please be patient as this may take a while.'))
        # 'hosts' is defined as 'all' in the playbook yml code, but inventory
        # filtering will subset that down to the hosts in playbook.hosts.
        self.playbook.hosts = hostnames
        self.playbook.services = servicenames
        self.playbook.extra_vars = 'kolla_action=stop'
        if self.playbook.verbose_level <= 1:
            self.playbook.print_output = False
        job = self.playbook.run()
        return job

    def precheck(self, hostnames=[], servicenames=[]):
        '''run check playbooks on a set of hosts'''

        # check that password file has no empty password values
        empty_keys = get_empty_password_values()
        if empty_keys:
            raise InvalidConfiguration(
                u._('password check failed. There are empty password values '
                    'in {etc}passwords.yml. '
                    'Please run kolla-cli password init or '
                    'kolla-cli password set(key) to correct them. '
                    '\nEmpty passwords: '
                    '{keys}').format(etc=get_kolla_etc(), keys=empty_keys))

        # define 'hosts' to be all, but inventory filtering will subset
        # that down to the hosts in playbook.hosts.
        self.playbook.hosts = hostnames
        self.playbook.services = servicenames
        self.playbook.extra_vars = 'kolla_action=precheck'
        self.playbook.print_output = True
        job = self.playbook.run()
        return job

    def pull(self, hostnames=[], servicenames=[]):
        '''run pull action against all hosts'''

        self.playbook.hosts = hostnames
        self.playbook.services = servicenames
        self.playbook.extra_vars = 'kolla_action=pull'
        self.playbook.print_output = True
        job = self.playbook.run()
        return job

    def upgrade(self, hostnames=[], servicenames=[]):
        '''Upgrades existing OpenStack Environment.'''

        self.playbook.hosts = hostnames
        self.playbook.services = servicenames
        self.playbook.extra_vars = 'kolla_action=upgrade'
        self.playbook.print_output = True
        job = self.playbook.run()
        return job

    def genconfig(self, hostnames=[], servicenames=[]):
        '''Generate configuration files for enabled OpenStack services'''

        self.playbook.hosts = hostnames
        self.playbook.services = servicenames
        self.playbook.extra_vars = 'kolla_action=config'
        self.playbook.print_output = True
        job = self.playbook.run()
        return job

    def check(self, hostnames=[], servicenames=[]):
        '''Do post-deployment smoke tests.'''

        self.playbook.hosts = hostnames
        self.playbook.services = servicenames
        self.playbook.extra_vars = 'kolla_action=check'
        self.playbook.print_output = True
        job = self.playbook.run()
        return job

    def _run_deploy_rules(self, playbook):
        properties = AnsibleProperties()
        inventory = Inventory.load()

        # cannot have both groups and hosts
        if playbook.hosts and playbook.groups:
            raise InvalidArgument(
                u._('Hosts and Groups arguments cannot '
                    'both be present at the same time.'))

        # verify that all services exists
        if playbook.services:
            for service in playbook.services:
                valid_service = inventory.get_service(service)
                if not valid_service:
                    raise NotInInventory(u._('Service'), service)

        # check that every group with enabled services
        # has hosts associated to it
        group_services = inventory.get_group_services()
        failed_groups = []
        failed_services = []
        if group_services:
            for (groupname, servicenames) in group_services.items():
                group = inventory.get_group(groupname)
                hosts = group.get_hostnames()

                group_needs_host = False
                if not hosts:
                    for servicename in servicenames:
                        if self._is_service_enabled(servicename,
                                                    inventory,
                                                    properties):
                            group_needs_host = True
                            failed_services.append(servicename)
                    if group_needs_host:
                        failed_groups.append(groupname)

            if len(failed_groups) > 0:
                raise InvalidConfiguration(
                    u._('Deploy failed. '
                        'Groups: {groups} with enabled '
                        'services : {services} '
                        'have no associated hosts')
                    .format(groups=failed_groups, services=failed_services))

    def _is_service_enabled(self, servicename, inventory, properties):
        service = inventory.get_service(servicename)
        if service is not None:
            enabled_property = 'enable_' + servicename.replace('-', '_')
            is_enabled = properties.get_property_value(enabled_property)
            if type(is_enabled) is str:
                is_enabled = is_string_true(is_enabled)
        return is_enabled



# Copyright (c) 2018 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from kolla_cli.tests.unit.common import KollaCliUnitTest


class TestUnit(KollaCliUnitTest):
    @mock.patch('kolla_cli.api.control_plane.ControlPlaneApi.upgrade')
    @mock.patch('kolla_cli.common.ansible.job.AnsibleJob.get_status')
    @mock.patch('kolla_cli.shell.KollaCli._is_inventory_present',
                return_value=True)
    def test_upgrade(self, _, mock_get_status, mock_upgrade):
        mock_get_status.return_value = 0
        mock_upgrade.return_value = self.get_fake_job()
        ret = self.run_cli_command('action upgrade')
        self.assertEqual(ret, 0)
        mock_upgrade.assert_called_once_with(1, [], [])

    @mock.patch('kolla_cli.api.control_plane.ControlPlaneApi.upgrade')
    @mock.patch('kolla_cli.common.ansible.job.AnsibleJob.get_status')
    @mock.patch('kolla_cli.shell.KollaCli._is_inventory_present',
                return_value=True)
    def test_upgrade_with_hosts(self, _, mock_get_status, mock_upgrade):
        mock_get_status.return_value = 0
        mock_upgrade.return_value = self.get_fake_job()
        hostnames = ['host1', 'host2']
        ret = self.run_cli_command(
            'action upgrade --hosts {hosts}'.format(
                hosts=','.join(hostnames)))
        self.assertEqual(ret, 0)
        mock_upgrade.assert_called_once_with(1, hostnames, [])

    @mock.patch('kolla_cli.api.control_plane.ControlPlaneApi.upgrade')
    @mock.patch('kolla_cli.common.ansible.job.AnsibleJob.get_status')
    @mock.patch('kolla_cli.shell.KollaCli._is_inventory_present',
                return_value=True)
    def test_upgrade_with_services(self, _, mock_get_status, mock_upgrade):
        mock_get_status.return_value = 0
        mock_upgrade.return_value = self.get_fake_job()
        services = ['service1', 'service2']
        ret = self.run_cli_command(
            'action upgrade --service {services}'.format(
                services=','.join(services)))
        self.assertEqual(ret, 0)
        mock_upgrade.assert_called_once_with(1, [], services)

    @mock.patch('kolla_cli.api.control_plane.ControlPlaneApi.upgrade')
    @mock.patch('kolla_cli.common.ansible.job.AnsibleJob.get_status')
    @mock.patch('kolla_cli.shell.KollaCli._is_inventory_present',
                return_value=True)
    def test_upgrade_with_hosts_and_services(self, _, mock_get_status,
                                             mock_upgrade):
        mock_get_status.return_value = 0
        mock_upgrade.return_value = self.get_fake_job()
        hostnames = ['host1', 'host2']
        services = ['service1', 'service2']
        ret = self.run_cli_command(
            'action upgrade --hosts {hosts} --service {services}'.format(
                hosts=','.join(hostnames), services=','.join(services)))
        self.assertEqual(ret, 0)
        mock_upgrade.assert_called_once_with(1, hostnames, services)



# Copyright (c) 2018 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import testtools

from kolla_cli.api.group import Group
from kolla_cli.api.host import Host
from kolla_cli.api.service import Service
from kolla_cli.common.ansible.job import AnsibleJob
from kolla_cli import shell


class KollaCliUnitTest(testtools.TestCase):

    def run_cli_command(self, command_string):
        # return 0 if command succeeded, non-0 if failed
        args = command_string.split()
        return shell.main(args)

    def get_fake_job(self):
        return AnsibleJob(None, None, None, None)

    def get_fake_host(self, hostname='foo'):
        return Host(hostname)

    def get_fake_group(self, groupname='group1', servicenames=[],
                       hostnames=[]):
        return Group(groupname, servicenames, hostnames)

    def get_fake_service(self, servicename='service1', parentnames=[],
                         childnames=[], groupnames=[]):
        return Service(servicename, parentnames, childnames, groupnames)



# Copyright(c) 2016, Oracle and/or its affiliates.  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from kolla_cli.api.exceptions import FailedOperation
from kolla_cli.common import utils
import kolla_cli.i18n as u

PWDS_FILENAME = 'passwords.yml'


def set_password(pwd_key, pwd_value):
    """set a password value

    If the password name exists, it will be changed.
    If it doesn't exist, a new password will be added.
    """
    value_switch = '-v'
    if not pwd_value:
        pwd_value = ''
        value_switch = ''
    cmd = '%s -k \'%s\' %s \'%s\'' % (_get_cmd_prefix(), pwd_key, value_switch,
                                      pwd_value)
    err_msg, output = utils.run_cmd(cmd, print_output=False)
    if err_msg:
        raise FailedOperation(
            u._('Password set failed. {error} {message}')
            .format(error=err_msg, message=output))


def set_password_sshkey(pwd_key, private_key, public_key):
    cmd = '%s -k \'%s\' -r \'%s\' -u \'%s\'' % (_get_cmd_prefix(), pwd_key,
                                                private_key, public_key)
    err_msg, output = utils.run_cmd(cmd, print_output=False)
    if err_msg:
        raise FailedOperation(
            u._('Password ssh key set failed. {error} {message}')
            .format(error=err_msg, message=output))


def clear_password(pwd_key):
    """clear a password

    if the password exists, it will be removed from the passwords file
    """
    cmd = '%s -k \'%s\' -c' % (_get_cmd_prefix(), pwd_key)
    err_msg, output = utils.run_cmd(cmd, print_output=False)
    if err_msg:
        raise FailedOperation('%s %s' % (err_msg, output))


def get_password_names():
    """return a list of password names"""
    cmd = '%s -l' % (_get_cmd_prefix())
    err_msg, output = utils.run_cmd(cmd, print_output=False)
    if err_msg:
        raise FailedOperation('%s %s' % (err_msg, output))

    pwd_names = []
    if output and ',' in output:
        pwd_names = output.strip().split(',')
    return pwd_names


def get_empty_password_values():
    cmd = '%s -e' % (_get_cmd_prefix())
    err_msg, output = utils.run_cmd(cmd, print_output=False)
    # output of this command is a comma separated string of password keys
    # that have empty values.
    if err_msg:
        raise FailedOperation('%s %s' % (err_msg, output))

    empty_keys = []
    if output:
        # password keys exist that have no values
        empty_keys = output.strip().split(',')
    return empty_keys


def init_passwords():
    # init empty passwords & ssh keys to auto-gen'd values
    cmd = '%s -i' % (_get_cmd_prefix())
    err_msg, output = utils.run_cmd(cmd, print_output=False)
    if err_msg:
        raise FailedOperation('%s %s' % (err_msg, output))


def _get_cmd_prefix():
    actions_path = utils.get_kolla_actions_path()
    pwd_file_path = os.path.join(utils.get_kolla_etc(),
                                 PWDS_FILENAME)
    prefix = ('%s password -p %s '
              % (actions_path, pwd_file_path))
    return prefix



# Copyright(c) 2016, Oracle and/or its affiliates.  All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#
from copy import copy


class Service(object):
    class_version = 1

    def __init__(self, name):
        self.name = name
        self._parentnames = []
        self._childnames = []
        self._groupnames = []
        self._supported = True
        self._vars = {}
        self.version = self.__class__.class_version

    def upgrade(self):
        pass

    def add_parentname(self, parentname):
        if parentname is not None and parentname not in self._parentnames:
            self._parentnames.append(parentname)

    def remove_parentname(self, parentname):
        if parentname in self._parentnames:
            self._parentnames.remove(parentname)

    def get_parentnames(self):
        return copy(self._parentnames)

    def add_groupname(self, groupname):
        if groupname is not None and groupname not in self._groupnames:
            self._groupnames.append(groupname)

    def remove_groupname(self, groupname):
        if groupname in self._groupnames:
            self._groupnames.remove(groupname)

    def set_groupnames(self, groupnames):
        self._groupnames = groupnames

    def get_groupnames(self):
        return copy(self._groupnames)

    def get_childnames(self):
        return copy(self._childnames)

    def add_childname(self, childname):
        if childname not in self._childnames:
            self._childnames.append(childname)

    def remove_childname(self, childname):
        if childname in self._childnames:
            self._childnames.remove(childname)

    def set_supported(self, supported):
        self._supported = supported

    def is_supported(self):
        return self._supported

    def get_vars(self):
        return self._vars.copy()



# Copyright(c) 2016, Oracle and/or its affiliates.  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import traceback

from cliff.command import Command

from kolla_cli.api.client import ClientApi
from kolla_cli.commands.exceptions import CommandError
import kolla_cli.i18n as u

LOG = logging.getLogger(__name__)
CLIENT = ClientApi()


class Setdeploy(Command):
    """Set deploy mode to either local or remote.

    Local indicates that the openstack deployment will be
    to the local host. Remote means that the deployment is
    on remote hosts.
    """
    def get_parser(self, prog_name):
        parser = super(Setdeploy, self).get_parser(prog_name)
        parser.add_argument('mode', metavar='<mode>',
                            help=u._('mode=<local, remote>'))
        return parser

    def take_action(self, parsed_args):
        try:
            mode = parsed_args.mode.strip()
            remote_flag = True
            if mode == 'local':
                remote_flag = False
                LOG.info(u._('Please note that local mode is not supported '
                             'and should never be used in production '
                             'environments.'))
            elif mode != 'remote':
                raise CommandError(
                    u._('Invalid deploy mode. Mode must be '
                        'either "local" or "remote".'))
            CLIENT.set_deploy_mode(remote_flag)
        except CommandError as e:
            raise e
        except Exception:
            raise Exception(traceback.format_exc())


class Getdeploy(Command):
    """get deploy mode from either local or remote.

    Local indicates that the openstack deployment will be
    to the local host. Remote means that the deployment is
    on remote hosts.
    """

    def take_action(self, parsed_args):
        try:
            mode = CLIENT.get_deploy_mode()
            return mode
        except CommandError as e:
            raise e
        except Exception:
            raise Exception(traceback.format_exc())



#!/usr/bin/env python
# Copyright(c) 2016, Oracle and/or its affiliates.  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import getopt
import os
import signal
import subprocess
import sys
import yaml

from kolla_cli.common.utils import change_password
from kolla_cli.common.utils import clear_all_passwords
from kolla_cli.common.utils import get_kolla_ansible_home
from kolla_cli.common.utils import get_kolla_cli_etc
from kolla_cli.common.utils import get_kolla_etc


def _init_keys(path):
    cmd = 'kolla-genpwd'
    if not os.path.exists(path):
        raise Exception('The path %s does not exist' % path)

    cmd = ' '.join((cmd, '-p', path))
    (_, err) = subprocess.Popen(cmd, shell=True,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE).communicate()
    if err:
        raise Exception('Error running %s: %s' % (cmd, err))


def _get_empty_keys(path):
    """get empty keys

    print string with keys that have empty pwd values
    """
    ok_empty = ['docker_registry_password']
    empty_keys = ''
    with open(path, 'r') as f:
        pwd_data = f.read()
    pwds = yaml.safe_load(pwd_data)
    comma = ''
    if pwds:
        for pwd_key, pwd_val in pwds.items():
            is_empty = False
            if not pwd_val and pwd_key not in ok_empty:
                is_empty = True
            elif isinstance(pwd_val, dict):
                if not pwd_val.get('private_key', None):
                    is_empty = True
                elif not pwd_val.get('public_key', None):
                    is_empty = True
            if is_empty:
                empty_keys = ''.join([empty_keys, comma, pwd_key])
                comma = ','
    if empty_keys:
        print(empty_keys)


def _print_pwd_keys(path):
    keys_str = ''
    prefix = ''
    with open(path, 'r') as f:
        pwd_data = f.read()
    pwds = yaml.safe_load(pwd_data)
    if pwds:
        for pwd_key in pwds.keys():
            keys_str = ''.join([keys_str, prefix, pwd_key])
            prefix = ','
    print(keys_str)


def _password_cmd(argv):
    """password command

    args for password command:
      -p path              # path to passwords.yaml
      -k key               # key of password
      -v value             # value of password (if not ssh keys)
      -r private key value # ssh private key
      -u public key value  # ssh public key
      -c                   # flag to clear the password
      -l                   # print to stdout a csv string of the existing keys
      -e                   # get keys of passwords with empty values
      -i                   # init empty keys and ssh keys
    """
    opts, _ = getopt.getopt(argv[2:], 'p:k:v:r:u:clei')
    path = ''
    pwd_key = ''
    pwd_value = ''
    pwd_ssh_private = ''
    pwd_ssh_public = ''
    clear_flag = False
    list_flag = False
    empty_flag = False
    init_flag = False
    for opt, arg in opts:
        if opt == '-p':
            path = arg
        elif opt == '-k':
            pwd_key = arg
        elif opt == '-v':
            pwd_value = arg
        elif opt == '-r':
            pwd_ssh_private = arg.replace('"', '')
        elif opt == '-u':
            pwd_ssh_public = arg.replace('"', '')
        elif opt == '-c':
            clear_flag = True
        elif opt == '-l':
            list_flag = True
        elif opt == '-e':
            empty_flag = True
        elif opt == '-i':
            init_flag = True
    if init_flag:
        # init empty keys
        _init_keys(path)
    elif list_flag:
        # print the password keys
        _print_pwd_keys(path)
    elif empty_flag:
        # get empty passwords
        _get_empty_keys(path)
    else:
        # edit/clear a password
        change_password(path, pwd_key, pvalue=pwd_value,
                        private_key=pwd_ssh_private,
                        public_key=pwd_ssh_public, clear=clear_flag)


def _job_cmd(argv):
    """jobs command

    args for job command
      -t       # terminate action
      -p pid   # process pid
    """
    opts, _ = getopt.getopt(argv[2:], 'tp:')
    pid = None
    term_flag = False
    for opt, arg in opts:
        if opt == '-p':
            pid = arg
        elif opt == '-t':
            term_flag = True

    if term_flag:
        try:
            os.kill(int(pid), signal.SIGKILL)
        except Exception as e:
            raise Exception('%s, pid %s' % (str(e), pid))


def _config_reset_cmd():
    """config_reset command

    args for config_reset command
    - none
    """
    kolla_etc = get_kolla_etc()
    kolla_home = get_kolla_ansible_home()
    kollacli_etc = get_kolla_cli_etc()

    group_vars_path = os.path.join(kolla_home, 'ansible/group_vars')
    host_vars_path = os.path.join(kolla_home, 'ansible/host_vars')
    globals_path = os.path.join(group_vars_path, '__GLOBAL__')
    inventory_path = os.path.join(kollacli_etc, 'ansible/inventory.json')

    # truncate global property and inventory files
    with open(globals_path, 'w') as globals_file:
        globals_file.truncate()

    with open(inventory_path, 'w') as inventory_file:
        inventory_file.truncate()

    # clear all passwords
    clear_all_passwords()

    # nuke all files under the kolla etc base, skipping everything
    # in the kolla-cli directory and the globals.yml and passwords.yml files
    for dir_path, dir_names, file_names in os.walk(kolla_etc, topdown=False):
        if 'kolla-cli' not in dir_path:
            for dir_name in dir_names:
                if dir_name != 'kolla-cli':
                    os.rmdir(os.path.join(dir_path, dir_name))

            for file_name in file_names:
                if file_name == 'passwords.yml' or file_name == 'globals.yml':
                    continue
                os.remove(os.path.join(dir_path, file_name))

    # nuke all property files under the kolla-ansible base other than
    # all.yml and the global property file which we truncate above
    for dir_path, _, file_names in os.walk(group_vars_path):
        for file_name in file_names:
            if (file_name != '__GLOBAL__' and
               file_name != 'all.yml'):
                os.remove(os.path.join(dir_path, file_name))

    for dir_path, _, file_names in os.walk(host_vars_path):
        for file_name in file_names:
            os.remove(os.path.join(dir_path, file_name))


def main():
    """perform actions on behalf of kolla user

    sys.argv:
    sys.argv[1]   # command

    Supported commands:
    - password
    - job
    - config_reset
    """
    if len(sys.argv) <= 1:
        raise Exception('Invalid number of parameters')

    command = sys.argv[1]
    if command == 'password':
        _password_cmd(sys.argv)
    elif command == 'job':
        _job_cmd(sys.argv)
    elif command == 'config_reset':
        _config_reset_cmd()
    else:
        raise Exception('Invalid command %s' % command)


if __name__ == '__main__':
    main()



# Copyright(c) 2016, Oracle and/or its affiliates.  All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from copy import copy


class SubService(object):
    class_version = 1

    def __init__(self, name):
        self.name = name

        self._groupnames = []
        self._parent_servicename = None

        self._vars = {}
        self.version = self.__class__.class_version

    def upgrade(self):
        pass

    def add_groupname(self, groupname):
        if groupname is not None and groupname not in self._groupnames:
            self._groupnames.append(groupname)

    def remove_groupname(self, groupname):
        if groupname in self._groupnames:
            self._groupnames.remove(groupname)

    def get_groupnames(self):
        return copy(self._groupnames)

    def set_parent_servicename(self, parent_svc_name):
        self._parent_servicename = parent_svc_name

    def get_parent_servicename(self):
        return copy(self._parent_servicename)

    def get_vars(self):
        return self.vars.copy()



# Copyright(c) 2017, Oracle and/or its affiliates.  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import sys

from kolla_cli.api.certificate import CertificateApi
from kolla_cli.api.config import ConfigApi
from kolla_cli.api.control_plane import ControlPlaneApi
from kolla_cli.api.group import GroupApi
from kolla_cli.api.host import HostApi
from kolla_cli.api.password import PasswordApi
from kolla_cli.api.properties import PropertyApi
from kolla_cli.api.service import ServiceApi
from kolla_cli.api.support import SupportApi

CONSOLE_MESSAGE_FORMAT = '%(message)s'

# TODO(bmace) - API version should probably be stored somewhere else
VERSION = '0.1'


class ClientApi(
        CertificateApi,
        ConfigApi,
        ControlPlaneApi,
        GroupApi,
        HostApi,
        PasswordApi,
        PropertyApi,
        ServiceApi,
        SupportApi,
        ):
    """Client API Notes

    Objects returned by the API contain a local copy of the information
    in the datastore. While changes made to the local copy will be
    reflected in the local object, changes made to the datastore
    from other objects will not be reflected in this local copy. The
    object will need to be re-fetched from the datastore to reflect
    the updates.
    """

    @staticmethod
    def get_version():
        # type: () -> str
        return VERSION

    @staticmethod
    def enable_console_logging(level, enable=True):
        # type: (int, bool) -> None
        """enable/disable console logging for the api

        enable: True/False
        level: logging.INFO, logging.DEBUG, logging.WARNING,
        logging.CRITICAL...
        """
        root_logger = logging.getLogger('')
        console = logging.StreamHandler(sys.stderr)
        if enable:
            console.setLevel(level)
            formatter = logging.Formatter(CONSOLE_MESSAGE_FORMAT)
            console.setFormatter(formatter)
            root_logger.addHandler(console)
        else:
            root_logger.removeHandler(console)
